import sys
from pathlib import Path
from tempfile import NamedTemporaryFile

import numpy as np
import polars as pl
from brainrender.actors import Points

from argclz import argument, str_tuple_type, validator, list_type
from neuralib.atlas.brainrender.core import BrainRenderCLI
from neuralib.atlas.brainrender.util import get_color
from neuralib.atlas.util import iter_source_coordinates, allen_to_brainrender_coord, as_coords_array

__all__ = ['RoiRenderCLI']


class RoiRenderCLI(BrainRenderCLI):
    """ROIs reconstruction with brainrender"""
    DESCRIPTION = 'ROIs reconstruction with brainrender'

    DEFAULT_ROI_COLORS = ['orange', 'magenta', 'dimgray']

    # ========== #
    # GROUP_ROIS #
    # ========== #

    GROUP_ROIS = 'ROI View Option'

    roi_region: str | tuple[str, ...] = argument(
        '--roi-region',
        metavar='NAME,...',
        type=str_tuple_type,
        default=(),
        group=GROUP_ROIS,
        help='only show rois in region(s)'
    )

    radius: float = argument(
        '--roi-radius',
        default=30,
        group=GROUP_ROIS,
        help='each roi radius'
    )

    roi_alpha: float = argument(
        '--roi-alpha',
        validator.float.in_range_closed(0, 1),
        default=1,
        group=GROUP_ROIS,
        help='region alpha value'
    )

    roi_colors: str | tuple[str, ...] = argument(
        '--roi-colors',
        metavar='COLOR,...',
        type=str_tuple_type,
        default=DEFAULT_ROI_COLORS,
        group=GROUP_ROIS,
        help='colors of rois per region'
    )

    region_col: str | None = argument(
        '--region-col',
        metavar='TREE..',
        default=None,
        group=GROUP_ROIS,
        help='if None, auto infer, and check the lowest merge level contain all the regions specified'
    )

    inverse_lut: bool = argument(
        '--inverse-lut',
        group=GROUP_ROIS,
        help='inverse right/left maps to ipsi/contra hemisphere look up table'
    )

    source_order: tuple[str, ...] | None = argument(
        '--source-order',
        metavar='SOURCE,...',
        type=str_tuple_type,
        default=None,
        group=GROUP_ROIS,
        help='source order to follow the roi_colors'
    )

    # ============== #
    # GROUP_ROI_LOAD #
    # ============== #

    GROUP_ROIS_LOAD = 'ROI load Option'

    classifier_file: Path | None = argument(
        '--classifier-file',
        validator.path.is_suffix('.csv').is_exists().optional(),
        metavar='FILE',
        type=Path,
        default=None,
        group=GROUP_ROIS_LOAD,
        help='csv output file from allenccf'
    )

    only_source: tuple[str, ...] | None = argument(
        '--only-source',
        metavar='SOURCE,...',
        type=str_tuple_type,
        help='only show the rois from the given source'
    )

    file: list[Path] | None = argument(
        '--file',
        validator.list().on_item(validator.path.is_suffix(['.csv', '.npy'])) | validator.optional(),
        metavar='FILE',
        type=list_type(Path),
        default=None,
        action='extend',
        group=GROUP_ROIS_LOAD,
        help="points file as 'npy' or 'csv'"
    )

    _need_close_file: list[NamedTemporaryFile] = []
    _point_file_list: list[str] = []

    def run(self):
        self.post_parsing()

        if not self._stop_render:
            self.render()
            self.render_output()

            if len(self._need_close_file) != 0:
                for f in self._need_close_file:
                    f.close()
                    Path(f.name).unlink(missing_ok=True)  # winOS

    def render(self):
        super().render()

        if self.classifier_file is not None:
            self._add_points_classifier_csv()

        if self.file is not None:
            self._add_points_generic_file()

        self._reconstruct_points_from_file()

    @property
    def _get_hemisphere_lut(self):
        if self.inverse_lut:
            return {'right': 'contra', 'left': 'ipsi', 'both': 'both'}
        else:
            return {'right': 'ipsi', 'left': 'contra', 'both': 'both'}

    def _add_points_classifier_csv(self):
        iter_coords = iter_source_coordinates(
            self.classifier_file,
            area=self.roi_region,
            source=self.only_source,
            region_col=self.region_col,
            hemisphere=self._get_hemisphere_lut[self.hemisphere],
            to_brainrender=True if self.coordinate_space == 'ccf' else False,
            source_order=self.source_order
        )

        ret = [sc.coordinates for sc in iter_coords]
        self._save_tempfile(ret)

    def _save_tempfile(self, rois_list: list[list[np.ndarray]] | list[np.ndarray]):
        for p in rois_list:
            if isinstance(p, np.ndarray):
                # os handle for NamedTemporaryFile, https://stackoverflow.com/a/23212515
                delete = False if sys.platform == 'win32' else True
                f = NamedTemporaryFile(prefix='.temp-run-3d-proj-', suffix='.npy', delete=delete)

                np.save(f, p)
                f.seek(0)
                self._point_file_list.append(f.name)
                self._need_close_file.append(f)

    def _add_points_generic_file(self):
        for it in self.file:
            self._point_file_list.append(str(it))

    def _reconstruct_points_from_file(self, error_while_empty: bool = False):
        for i, file in enumerate(self._point_file_list):
            # type handle
            if file.endswith('.npy'):
                data = np.load(file)
            elif file.endswith('.csv'):
                data = pl.read_csv(file)
                data = as_coords_array(data)
                if self.coordinate_space == 'ccf':
                    data = allen_to_brainrender_coord(data)
            else:
                raise ValueError('Unsupported file type')

            # check
            if data.ndim != 2:
                raise ValueError(f'wrong dimension: {data.shape}')

            # add points
            if data.size == 0:
                if error_while_empty:
                    raise ValueError('no points found')
                else:
                    self.logger.warn('no points found')
            else:
                if data.shape[1] == 3:
                    colors = get_color(i, self.roi_colors)
                    self.logger.info(f'Plot Rois File: {i}, {file}, {colors}')
                    points = Points(data, name='roi', colors=colors, alpha=self.roi_alpha, radius=self.radius, res=20)
                    self.scene.add(points)
                elif data.shape[1] == 4:  # TODO not test yet
                    k = data[:, 3].astype(int)
                    for t in np.unique(k):
                        self.scene.add(Points(
                            data[k == t, 0:3],
                            name='rois',
                            colors=get_color(t, self.roi_colors),
                            alpha=self.roi_alpha,
                            radius=self.radius
                        ))
                else:
                    raise ValueError(f'wrong shape: {data.shape}: {file}')


if __name__ == '__main__':
    RoiRenderCLI().main()



from typing import Union, Optional

from bokeh.model import Model
from bokeh.models import Div

from neuralib.dashboard import View
from .model import list_date

__all__ = ['AnimalFigureView']


class AnimalFigureView(View):
    FIGURE_TYPE = ["default", 'other', 'more']
    content_title: Div
    content_date: Div

    def __init__(self, animal: str = None, figure: str = None):
        self._animal: Union[None, str, BaseException] = animal
        self._figure: Optional[str] = figure

    @property
    def title(self) -> str:
        animal = self.current_animal or 'Unknown'
        return f'Animal {animal}'

    @property
    def current_animal(self) -> Optional[str]:
        if self._animal is None:
            try:
                self._animal = self.get_arg('animal')[0]
            except (KeyError, IOError) as e:
                self._animal = e

        if isinstance(self._animal, str):
            return self._animal
        else:
            return None

    @property
    def current_figure(self) -> str:
        if self._figure is None:
            try:
                self._figure = self.get_arg('figure')[0]
            except (KeyError, IOError):
                self._figure = 'default'
        return self._figure

    def setup(self) -> Model:
        self.content_title = Div(text=f"animal {self.current_animal} with figure {self.current_figure}")
        self.content_date = Div(text="Date:<ol>" + "\n".join([
            f"<li>{date}</li>" for date in list_date(self.current_animal)
        ]) + "</ol>")

        from bokeh.layouts import column
        return column(
            self.content_title,
            self.content_date,
        )



"""
Profile
=================

This module provides utilities for profiling and tracing execution of code.

It includes:

- **profile_test**: A context manager for profiling code execution using the
  built-in `cProfile` module. The profiler can dump the stats in various formats:
  - `.dat`: Raw profile dump.
  - `.txt`: Human-readable text output.
  - `.png`: Graphical representation using `gprof2dot` and Graphviz's `dot` tool.

- **trace_line**: A decorator that launches a background thread to trace and
  print the current line number being executed at regular intervals. This is
  especially useful when debugging long-running processes that might be
  terminated by the operating system without a traceback.

Usage Examples
--------------

Profiling Example:

.. code-block:: python

    with profile_test(enable=True, output_file='profile.txt'):
        # Code to profile
        do_something()

Trace Line Example:

.. code-block:: python

    @trace_line(interval=0.1)
    def long_running_function():
        # Some long-running process
        process_data()

    long_running_function()

"""
import functools
import sys
import threading
import time

from neuralib.util.verbose import fprint

__all__ = ['profile_test', 'trace_line']


class profile_test:
    def __init__(self, enable=False, output_file='profile.png'):
        self._enable = enable
        self._profile = None
        self._output_file = output_file

    def __enter__(self):
        if self._enable:
            import cProfile

            self._profile = cProfile.Profile()
            self._profile.enable()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._enable:
            self._profile.disable()

            if self._output_file.endswith('.dat'):
                self._profile.dump_stats(self._output_file)

            elif self._output_file.endswith('.txt'):
                import pstats

                stat = pstats.Stats(self._profile)

                with open(self._output_file, 'w') as f:
                    stat.stream = f
                    stat.print_stats()

            elif self._output_file.endswith('.png'):
                import subprocess

                f1 = self._output_file.replace('.png', '.dat')
                self._profile.dump_stats(f1)
                cmd_line = f'python -m gprof2dot -f pstats {f1} | dot -T png -o {self._output_file}'

                with subprocess.Popen(['bash', '-c', cmd_line]) as proc:
                    proc.wait()


def trace_line(f=None, *, interval: float = 0.1):
    """
    Decorator for trace line number in real time using threading.

    **Use Case**

    - OS directly kill job without traceback

    :param f: Decorated function
    :param interval: Time interval for trace (in second).
    """

    def _decorator(f):

        @functools.wraps(f)
        def _wrapper(*args, **kwargs):
            done = False
            _file = None
            _lineno = None
            _prev_time = None

            def update():
                nonlocal done
                nonlocal _file
                nonlocal _lineno
                nonlocal _prev_time

                while not done:
                    time.sleep(interval)

                    # noinspection PyUnresolvedReferences,PyProtectedMember
                    info = sys._current_frames()[caller_thread]

                    if info.f_code.co_filename != _file or info.f_lineno != _lineno:

                        current_time = time.time()
                        if _prev_time is not None:
                            elapsed_time = current_time - _prev_time
                        else:
                            elapsed_time = 0  # First iteration

                        _file = info.f_code.co_filename
                        _lineno = info.f_lineno
                        _prev_time = current_time

                        fprint(f'File: {_file} in line {_lineno} - Time elapsed: {elapsed_time:.3f} sec',
                               vtype='debug',
                               flush=True)

            thread = threading.Thread(target=update)
            caller_thread = threading.current_thread().ident
            thread.start()

            try:
                result = f(*args, **kwargs)
            finally:
                done = True  # Signal to stop the memory monitoring thread
                thread.join()  # Ensure the thread finishes

            return result

        return _wrapper

    if f is None:
        return _decorator
    else:
        return _decorator(f)



from __future__ import annotations

from pathlib import Path

import numpy as np
from typing_extensions import Self

from neuralib.typing import PathLike

__all__ = [
    'read_cellpose',
    'cellpose_point_roi_helper',
    'CellposeSegmentation'
]


def read_cellpose(file: PathLike) -> CellposeSegmentation:
    """
    Read a cellpose segmentation result file

    :param file: cellpose segmentation result ``.npy`` file
    :return: :class:`CellposeSegmentation`
    """
    return CellposeSegmentation.load(file)


def cellpose_point_roi_helper(file: PathLike, output: PathLike) -> None:
    """
    Read a cellpose segmentation result and convert the segmentation result to point coordinates

    :param file: cellpose segmentation result ``.npy`` file
    :param output: ``*.roi`` output file path
    """
    CellposeSegmentation.load(file).to_roi(output)


class CellposeSegmentation:
    """`Cellpose <https://github.com/MouseLand/cellpose>`_ segmentation results

    `Dimension parameters`:

        N = Number of segmented cells

        W = Image width

        H = Image height

    .. seealso::

        `Cellpose Native Doc <https://cellpose.readthedocs.io/en/latest/outputs.html#seg-npy-output>`_

    """

    def __init__(self, outlines, masks, chan_choose, ismanual, flows, diameter, filename):
        self._outlines = outlines
        self._masks = masks
        self._chan_choose = chan_choose
        self._is_manual = ismanual
        self._flows = flows
        self._diameter = diameter

        self._filename = filename

    @classmethod
    def load(cls, file: PathLike) -> Self:
        """
        Load a cellpose segmentation result

        :param file: cellpose segmentation result ``.npy`` file
        :return: :class:`CellposeSegmentation`
        """
        dat = np.load(file, allow_pickle=True).item()
        return cls(**dat)

    @property
    def n_segmentation(self) -> int:
        """number of segmented cells"""
        return len(self._is_manual)

    @property
    def width(self):
        """image width"""
        return self._outlines.shape[1]

    @property
    def height(self):
        """image height"""
        return self._outlines.shape[0]

    @property
    def filename(self) -> Path:
        """filepath of image"""
        return Path(self._filename)

    @property
    def outlines(self) -> np.ndarray:
        """outlines of ROIs (0 = NO outline; 1,2,… = outline labels). `Array[uint16, [H, W]]`"""
        return self._outlines

    @property
    def masks(self) -> np.ndarray:
        """each pixel in the image is assigned to an ROI (0 = NO ROI; 1,2,… = ROI labels). `Array[uint16, [H, W]]` """
        return self._masks

    @property
    def chan_choose(self) -> list[int]:
        """channels that you chose in GUI (0=gray/none, 1=red, 2=green, 3=blue)"""
        return self._chan_choose

    @property
    def flows(self) -> list[np.ndarray]:
        """
        flows[0] is XY flow in RGB

        flows[1] is the cell probability in range 0-255 instead of -10.0 to 10.0

        flows[2] is Z flow in range 0-255 (if it exists, otherwise zeros),

        flows[3] is [dY, dX, cellprob] (or [dZ, dY, dX, cellprob] for 3D), flows[4] is pixel destinations (for internal use)
        """
        return self._flows

    @property
    def diameter(self) -> float:
        """cell body diameter"""
        return self._diameter

    @property
    def is_manual(self) -> np.ndarray:
        """whether or not mask k was manually drawn or computed by the cellpose algorithm. `Array[bool, N]`"""
        return self._is_manual

    @property
    def nan_masks(self) -> np.ndarray:
        """value 0 in :attr:`CellposeSegmentation.masks` to nan"""
        masks = self.masks.copy().astype(np.float_)
        masks[masks == 0] = np.nan

        return masks

    @property
    def nan_outlines(self) -> np.ndarray:
        """value 0 in :attr:`CellposeSegmentation.outlines` to nan"""
        outlines = self.outlines.copy().astype(np.float_)
        outlines[outlines == 0] = np.nan

        return outlines

    @property
    def points(self) -> np.ndarray:
        """Calculate center of each segmented area in XY pixel. `Array[int, [N, 2]]`"""
        centers = self._calculate_centers()
        return np.round(centers).astype(int)

    # noinspection PyTypeChecker
    def to_roi(self, output_file: PathLike):
        """
        Covert segmented roi to point roi, and save it as ``.roi`` for imageJ.

        :param output_file: ``*.roi`` output file path
        """
        from roifile import ImagejRoi, ROI_TYPE, ROI_OPTIONS

        if Path(output_file).suffix != '.roi':
            raise ValueError('output file must have .roi extension')

        points = np.fliplr(self.points)  # XY rotate in .roi format
        roi = ImagejRoi(
            roitype=ROI_TYPE.POINT,
            options=ROI_OPTIONS.PROMPT_BEFORE_DELETING | ROI_OPTIONS.SUB_PIXEL_RESOLUTION,
            n_coordinates=self.points.shape[0],
            integer_coordinates=points,
            subpixel_coordinates=points
        )

        roi.tofile(output_file)

    def _calculate_centers(self):
        """calculate center of each segmented area in XY pixel"""
        labels = np.unique(self.masks)
        labels = labels[labels != 0]  # remove background

        n_neurons = len(labels)
        centers = np.zeros((n_neurons, 2))
        for i, label in enumerate(labels):
            segment_coords = np.argwhere(self.masks == label)
            center = segment_coords.mean(axis=0)
            centers[i] = center

        return centers



from pathlib import Path

__all__ = [
    'NEUROLIB_CACHE_DIRECTORY',
    'NEUROLIB_DATASET_DIRECTORY',
    'ATLAS_CACHE_DIRECTORY',
    'CASCADE_MODEL_CACHE_DIRECTORY'
]

CACHE_DIRECTORY = Path.home() / '.cache'
NEUROLIB_CACHE_DIRECTORY = CACHE_DIRECTORY / 'neuralib'
NEUROLIB_DATASET_DIRECTORY = NEUROLIB_CACHE_DIRECTORY / 'dataset'

# ATLAS
ATLAS_CACHE_DIRECTORY = NEUROLIB_CACHE_DIRECTORY / 'atlas'

# MODEL
CASCADE_MODEL_CACHE_DIRECTORY = NEUROLIB_CACHE_DIRECTORY / 'cascade'



from typing import Literal

from neuralib.typing import Series
from .typing import Area, TreeLevel

__all__ = [
    'NUM_MERGE_LAYER',
    'DEFAULT_FAMILY_DICT',
    'ALLEN_FAMILY_TYPE',
    'merge_until_level'
]

NUM_MERGE_LAYER = 5

# =========================== #
# LEVEL 4: Mainly Merge layer #
# =========================== #

MERGE_LEVEL_LAYER: TreeLevel = 4
MERGE_REGION_LV4 = {
    # VS
    'VL': ['SEZ', 'chpl'],
    'V4': ['V4r'],

    # CB
    'AN': ['ANcr1', 'ANcr2'],
    'CENT': ['CENT*'],
    'CUL': ['CUL*'],

    # HPF
    'ENTl': ['ENTl*'],
    'ENTm': ['ENTm*'],

    # Isocortex
    'ACAd': ['ACAd*'],
    'ACAv': ['ACAv*'],
    'AId': ['AId*'],
    'AIp': ['AIp*'],
    'AIv': ['AIv*'],
    'AUDd': ['AUDd*'],
    'AUDp': ['AUDp*'],
    'AUDpo': ['AUDpo*'],
    'AUDv': ['AUDv*'],
    'ECT': ['ECT*'],
    'FRP': ['FRP*'],
    'GU': ['GU*'],
    'ILA': ['ILA*'],
    'MOp': ['MOp*'],
    'MOs': ['MOs*'],
    'ORBl': ['ORBl*'],
    'ORBm': ['ORBm*'],
    'ORBvl': ['ORBvl*'],
    'PERI': ['PERI*'],
    'PL': ['PL*'],
    'VISa': ['VISa1', 'VISa2/3', 'VISa4', 'VISa5', 'VISa6a', 'VISa6b'],  # NOTE PTLp child
    'VISrl': ['VISrl*'],
    'RSPagl': ['RSPagl*'],
    'RSPd': ['RSPd*'],
    'RSPv': ['RSPv*'],
    'SSs-bfd': ['SSs-bfd*'],
    'SSp-ll': ['SSp-ll*'],
    'SSp-m': ['SSp-m*'],
    'SSp-n': ['SSp-n*'],
    'SSp-tr': ['SSp-tr*'],
    'SSp-ul': ['SSp-ul*'],
    'SSp-un': ['SSp-un*'],
    'SSs': ['SSs*'],
    'TEa': ['TEa*'],
    'VISal': ['VISal*'],
    'VISam': ['VISam*'],
    'VISl': ['VISl1', 'VISl2/3', 'VISl4', 'VISl5', 'VISl6a', 'VISl6b'],  # differentiate the VISli
    'VISli': ['VISli*'],
    'VISp': ['VISp1', 'VISp2/3', 'VISp4', 'VISp5', 'VISp6a', 'VISp6b'],  # differentiate  VISpl, VISpm
    'VISpl': ['VISpl*'],
    'VISpm': ['VISpm*'],
    'VISpor': ['VISpor*'],
    'VISC': ['VISC*'],

    # OLF
    'NLOT': ['NLOT*']

}

# ============================== #
# LEVEL 3: Mainly merge DV ML AP #
# ============================== #

MERGE_LEVEL_DVMLAP: TreeLevel = 3
MERGE_REGION_LV3 = {
    # HB
    'AMB': ['AMB*'],
    'LRN': ['LRN*'],
    'MDRN': ['MDRN*'],
    'PGRN': ['PGRN*'],
    'SOC': ['POR', 'SOC*'],

    # HY
    'MM': ['MMd', 'MMl', 'MMm', 'MMme', 'MMp',
           'Mmd', 'Mml', 'Mmm', 'Mmme', 'Mmp'],  # diff in ccf
    'TM': ['TM*'],

    # TH
    'AM': ['AM*'],
    'LGd': ['LGd-*'],
    'MG': ['MG*'],
    'SPF': ['SPF*'],
    'VP': ['VPL', 'VPLpc', 'VPM', 'VPMpc'],

    # HPF
    'ENT': ['ENT*'],
    'DG': ['DG*'],

    # Isocortex
    'ACA': ['ACA*'],
    'AI': ['AId', 'AIp', 'AIv'],
    'AUD': ['AUD*'],
    'ORB': ['ORBl', 'ORBm', 'ORBvl'],
    'SSp': ['SSp*'],
    'AOB': ['AOB*'],
    'COA': ['COA*'],
    'TT': ['TT*'],

    # CNU
    'PALc': ['BAC', 'BST'],
    'PALd': ['GPe', 'GPi'],
    'MSC': ['MS', 'NDB'],
    'PALv': ['MA', 'SI'],
    'LS': ['LSc', 'LSr', 'LSv'],
    'CEA': ['CEA*'],

    # MB
    'SCm': ['SCdg', 'SCdw', 'SCig', 'SCiw'],
    'IC': ['IC*'],
    'SCs': ['SCop', 'SCsg', 'SCzo'],

    # CTXsp
    'BLA': ['BLA*'],
    'BMA': ['BMA*'],
    'EP': ['EP*'],

}

# ========================= #
# LEVEL 2: Customized Merge #
# ========================= #

MERGE_LEVEL_C2: TreeLevel = 2
MERGE_REGION_LV2 = {

    # HPF
    'CA': ['CA1', 'CA2', 'CA3'],
    # 'SUB': ['ProS'],  # might not correct, converge due to the area not in cellatlas (Prosubiculum)

    # Isocortex
    'SS': ['SS*'],
    'VIS': ['VISal', 'VISam', 'VISl', 'VISli', 'VISp', 'VISpl', 'VISpm', 'VISpor'],
    'PTLp': ['VISa', 'VISrl'],
    'MO': ['MOp', 'MOs'],
    'RSP': ['RSP*'],

    # TH
    'ATN': ['AD', 'AM', 'AV', 'IAD', 'IAM', 'LD'],
    'EPI': ['LH', 'MH'],
    'GENv': ['IGL', 'IntG', 'LGv', 'SubG'],
    'ILM': ['CL', 'CM', 'PCN', 'PF', 'PIL', 'RH'],
    'LAT': ['Eth', 'LP', 'PO', 'POL', 'SGN'],
    'MED': ['IMD', 'MD', 'PR', 'SMT'],
    'MTN': ['PT', 'PVT', 'RE', 'Xi'],
    'GENd': ['LGd', 'MG'],
    'VENT': ['PoT', 'VAL', 'VM', 'VP'],

    # CNU
    'PALm': ['MSC', 'TRS'],
    'LSX': ['LS', 'SF', 'SH'],
    'STRd': ['CP'],
    'STRv': ['ACB', 'FS', 'OT'],
    'sAMY': ['AAA', 'BA', 'CEA', 'IA', 'MEA'],

    # HY
    'MBO': ['LM', 'MM', 'SUM', 'TM'],
    'ZI': ['FF'],

    # MB
    'IPN': ['IPA', 'IPC', 'IPDL', 'IPDM', 'IPI', 'IPL', 'IPR', 'IPRL'],

    # HB
    'PHY': ['NR', 'PRP'],
    'VNC': ['LAV', 'MV', 'SPIV', 'SUV'],
    'CN': ['DCO', 'VCO'],
    'DCN': ['CU', 'GR'],
    'PB': ['KF'],

    # CB
    'HEM': ['AN', 'COPY', 'FL', 'PFL', 'PRM', 'SIM'],
    'VERM': ['CENT', 'CUL', 'DEC', 'FOTU', 'LING', 'NOD', 'PYR', 'UVU'],

}

# ========================= #
# LEVEL 1: Customized Merge #
# ========================= #

MERGE_LEVEL_C1: TreeLevel = 1
MERGE_REGION_LV1 = {
    # VS
    'VS': ['AQ', 'V3', 'V4', 'VL'],

    # OLF
    'OLF': ['AOB', 'AON', 'COA', 'DP', 'MOB', 'NLOT', 'PAA', 'PIR', 'TR', 'TT'],

    # HPF
    'HIP': ['CA', 'DG', 'FC', 'IG'],
    'RHP': ['APr', 'ENT', 'HATA', 'PAR', 'POST', 'PRE', 'ProS', 'SUB'],

    # CNU
    'PAL': ['PALc', 'PALd', 'PALm', 'PALv'],
    'STR': ['LSX', 'STRd', 'STRv', 'sAMY'],

    # HY
    'LZ': ['LHA', 'LPO', 'PST', 'PSTN', 'PeF', 'RCH', 'STN', 'TU', 'ZI'],
    'MEZ': ['AHN', 'MBO', 'MPN', 'PH', 'PMd', 'PMv', 'PVHd', 'VMH'],
    'PVR': ['ADP', 'AVP', 'AVPV', 'DMH', 'MEPO', 'MPO', 'OV', 'PD', 'PS', 'PVp', 'PVpo', 'SBPV', 'SCH', 'SFO', 'VLPO',
            'VMPO'],
    'PVZ': ['ARH', 'ASO', 'PVH', 'PVa', 'PVi', 'SO'],

    # MB
    'PAG': ['INC', 'ND', 'PRC', 'Su3'],
    'PRT': ['APN', 'MPT', 'NOT', 'NPC', 'OP', 'PPT', 'RPF'],
    'RAmb': ['CLI', 'DR', 'IF', 'IPN', 'RL']

}

# ========================= #
# LEVEL 0: Customized Merge #
# ========================= #

MERGE_LEVEL_C0: TreeLevel = 0
MERGE_REGION_LV0 = {
    # HB
    'MY-mot': ['ACVII', 'AMB', 'DMX', 'GRN', 'ICB', 'IO', 'IRN', 'ISN', 'LIN', 'LRN', 'MARN', 'MDRN', 'PARN', 'PAS',
               'PGRN', 'PHY', 'PPY', 'VI', 'VII', 'VNC', 'XII', 'x', 'y'],
    'MY-sat': ['RM', 'RO', 'RPA'],
    'MY-sen': ['AP', 'CN', 'DCN', 'ECU', 'NTB', 'NTS', 'PA5', 'SPVC', 'SPVI', 'SPVO'],
    'P-mot': ['Acs5', 'B', 'DTN', 'I5', 'P5', 'PC5', 'PCG', 'PDTg', 'PG', 'PRNc', 'SG', 'SUT', 'TRN', 'V'],
    'P-sat': ['CS', 'LC', 'LDT', 'NI', 'PRNr', 'RPO', 'SLC', 'SLD'],
    'P-sen': ['NLL', 'PB', 'PSV', 'SOC'],

    #
    'HPF': ['HIP', 'RHP'],

    #
    'HY': ['LZ', 'ME', 'MEZ', 'PVR', 'PVZ'],

    #
    'TH': ['ATN', 'EPI', 'GENv', 'ILM', 'LAT', 'MED', 'MTN', 'RT',  # DORpm
           'GENd', 'PP', 'SPA', 'SPF', 'VENT'],  # DORsm

    # MB
    'MBmot': ['AT', 'CUN', 'DT', 'EW', 'III', 'IV', 'LT', 'MA3', 'MRN', 'MT', 'PAG',
              'PN', 'PRT', 'Pa4', 'RN', 'RR', 'SCm', 'SNr', 'VTA', 'VTN'],
    'MBsen': ['IC', 'MEV', 'NB', 'PBG', 'SAG', 'SCO', 'SCs'],
    'MBsta': ['PPN', 'RAmb', 'SNc'],

    # CB
    'CBN': ['DN', 'FN', 'IP', 'VeCB'],
    'CBX': ['HEM', 'VERM']

}

# ==================================================================================== #
# Grey matter Family (i.e., Used after merging, assume already converge(merge) enough) #
# ==================================================================================== #

ALLEN_FAMILY_TYPE = Literal['HB', 'HY', 'TH', 'MB', 'CB', 'CTXpl', 'HPF', 'ISOCORTEX', 'OLF', 'CTXsp']
AllenFamilyType = tuple[Area, ...]

# VS
VS_FAMILY: AllenFamilyType = ('VS',)

# HB
HB_FAMILY: AllenFamilyType = ('MY-mot', 'MY-sat', 'MY-sen', 'P-mot', 'P-sat', 'P-sen')

# HY
HY_FAMILY: AllenFamilyType = ('HY',)

# TH
TH_FAMILY: AllenFamilyType = ('TH',)

# MB (`MB` initially show in ccf map)
MB_FAMILY: AllenFamilyType = ('MB', 'MBmot', 'MBsen', 'MBsta')

# CB
CB_FAMILY: AllenFamilyType = ('CBN', 'CBX')

# CNU
CNU_FAMILY: AllenFamilyType = ('PAL', 'STR')

# CTXpl
HPF_FAMILY: AllenFamilyType = ('HPF',)
ISOCORTEX_FAMILY: AllenFamilyType = (
    'ACA', 'AI', 'AUD', 'ECT', 'FRP', 'GU', 'ILA', 'MO', 'ORB', 'PERI', 'PL', 'PTLp', 'RSP',
    'SS', 'TEa', 'VIS', 'VISC'
)
OLF_FAMILY: AllenFamilyType = ('OLF',)
CTXPL_FAMILY = HPF_FAMILY + ISOCORTEX_FAMILY + OLF_FAMILY

# CTXsp
CTXSP_FAMILY: AllenFamilyType = ('CTXsp', 'BLA', 'BMA', 'CLA', 'EP', 'LA', 'PA')

DEFAULT_FAMILY_DICT: dict[str, AllenFamilyType] = dict(
    VS=VS_FAMILY,
    HB=HB_FAMILY,
    HY=HY_FAMILY,
    TH=TH_FAMILY,
    MB=MB_FAMILY,
    CB=CB_FAMILY,
    CNU=CNU_FAMILY,
    HPF=HPF_FAMILY,
    ISOCORTEX=ISOCORTEX_FAMILY,
    OLF=OLF_FAMILY,
    CTXSP=CTXSP_FAMILY
)


# ============== #
# Implementation #
# ============== #

def merge_area(ps: Series, region: dict[str, list[Area]]) -> list[str]:
    """
    merge the area name series to another level based on the ``region`` dict

    :param ps: pandas/polars series with area name
    :param region: ``MERGE_REGION_LV*``
    :return:
    """
    ret = []
    for a in ps:
        ret.append(_merge_area(a, region))

    return ret


def _merge_area(s: str, region: dict[str, list[Area]]) -> str:
    """

    :param s: individual `acronym` value
    :param region:
    :return:
    """
    for n, p in region.items():
        for pattern in p:
            if pattern.endswith('*'):
                if s.startswith(pattern[:-1]):
                    return n
            elif pattern == s:
                return n

    return s


def merge_until_level(ps: Series, level: TreeLevel) -> list[Area]:
    """
    merge the area until which `level`

    :param ps: pandas/polars series with area name
    :param level: level int
    :return: list of area
    """
    if level not in list(range(NUM_MERGE_LAYER)):
        raise ValueError(f'wrong level: {level}')

    if level <= MERGE_LEVEL_LAYER:
        ps = merge_area(ps, MERGE_REGION_LV4)
    if level <= MERGE_LEVEL_DVMLAP:
        ps = merge_area(ps, MERGE_REGION_LV3)
    if level <= MERGE_LEVEL_C2:
        ps = merge_area(ps, MERGE_REGION_LV2)
    if level <= MERGE_LEVEL_C1:
        ps = merge_area(ps, MERGE_REGION_LV1)
    if level <= MERGE_LEVEL_C0:
        ps = merge_area(ps, MERGE_REGION_LV0)

    return ps



from __future__ import annotations

from contextlib import contextmanager
from pathlib import Path
from typing import Any, Generator, final, Literal
from xml.etree.ElementTree import tostring

import aicspylibczi
import numpy as np
import xmltodict

from neuralib.typing import PathLike
from neuralib.util.unstable import unstable
from .core import AbstractScanner, SceneIdx, DimCode

__all__ = ['czi_file', 'CziScanner']


@contextmanager
def czi_file(filepath: PathLike) -> Generator[CziScanner, None, None]:
    """
    context manager for load a czi file

    :param filepath: czi file path
    :return: :class:`~CziScanner`
    """
    if Path(filepath).suffix != '.czi':
        raise ValueError('czi file only')

    scanner = CziScanner(filepath)
    try:
        yield scanner
    finally:
        scanner.close()


@final
class CziScanner(AbstractScanner):
    """CZI confocal image data"""

    _czi_file: aicspylibczi.CziFile | None
    _consistent_scan_configs: bool

    def __init__(self, filepath: PathLike):
        self._czi_file = aicspylibczi.CziFile(filepath)
        super().__init__(filepath)

    def _load_metadata(self) -> dict[str, Any]:
        dim = self._czi_file.get_dims_shape()
        first_scene_dims = dim[0]
        if 'S' in first_scene_dims and len(dim) == 1:
            self._n_scenes = first_scene_dims['S'][1]
            self._consistent_scan_configs = True
        elif len(dim) > 1:
            self._n_scenes = len(dim)
            self._consistent_scan_configs = len(set(tuple(sorted(d.items())) for d in dim)) == 1
        else:
            self._n_scenes = 1
            self._consistent_scan_configs = True

        xml_string = tostring(self._czi_file.meta, encoding='utf-8').decode('utf-8')
        return xmltodict.parse(xml_string)

    def close(self):
        if self._czi_file is not None:
            self._czi_file = None

    @property
    def czi_file(self) -> aicspylibczi.CziFile:
        """get ``aicspylibczi.CziFile`` object"""
        return self._czi_file

    @property
    def consistent_config(self) -> bool:
        """Checks if the CZI file have consistent scanning configs across scenes"""
        return self._consistent_scan_configs

    @property
    def is_mosaic(self) -> bool:
        """Checks if the CZI file is marked as mosaic by the reader."""
        return self._czi_file.is_mosaic()

    @property
    def n_scenes(self) -> int:
        """Total number of scenes (positions/series) in the file."""
        return self._n_scenes

    @property
    def dimcode(self) -> DimCode:
        return self.czi_file.dims

    def get_code(self, scene_idx: SceneIdx, code: Literal['S', 'C', 'Z', 'X', 'Y', 'M']) -> int:
        """
        Retrieves a specific shape value associated with a given scene index and
        access code from the internal czi_file data structure.

        :param scene_idx: The index representing a specific scene in the czi_file's dimensions.
        :param code: A string key for accessing a particular value related to the given scene index.
        :return: Returns the shape value corresponding to the scene index and ``code``
        """
        if scene_idx < 0 or scene_idx >= self.n_scenes:
            raise ValueError(f'invalid scene index:{scene_idx}')

        dim_shape = self.czi_file.get_dims_shape()
        if self.consistent_config:
            return dim_shape[0][code][1]
        else:
            return dim_shape[scene_idx][code][1]

    def get_tile_info(self, scene_idx: SceneIdx = 0) -> dict[str, Any] | None:
        """Extracts tile region information from metadata, if available.
        Returns None if not a tiled acquisition or info is missing"""
        tile_region_info = self.metadata \
            .get('ImageDocument', {}) \
            .get('Metadata', {}) \
            .get('Experiment', {}) \
            .get('ExperimentBlocks', {}) \
            .get('AcquisitionBlock', {}) \
            .get('SubDimensionSetups', {}) \
            .get('RegionsSetup', {}) \
            .get('SampleHolder', {}) \
            .get('TileRegions', {}) \
            .get('TileRegion')

        if tile_region_info is None:
            return None

        if isinstance(tile_region_info, list):
            if 0 <= scene_idx < len(tile_region_info):
                return tile_region_info[scene_idx]  # Return specific region dict
            else:
                return None
        elif isinstance(tile_region_info, dict):
            if scene_idx == 0:
                return tile_region_info
            else:
                return tile_region_info
        else:
            return None

    def get_channel_names(self, scene_idx: SceneIdx) -> list[str]:
        """Get the names of the fluorescence channels for a specific scene.
        (Implementation copied from czi_scanner_impl_v1 for completeness)
        """
        if not 0 <= scene_idx < self.n_scenes:
            raise IndexError(f"scene_idx {scene_idx} out of bounds for {self.n_scenes} scenes.")

        info = self.metadata.get('ImageDocument', {}).get('Metadata', {})
        channels_info = info.get('Information', {}).get('Image', {}).get('Dimensions', {}).get('Channels', {}).get('Channel')
        if isinstance(channels_info, list):
            names = [ch.get('@Name', f"Channel {i}") for i, ch in enumerate(channels_info)]
        elif isinstance(channels_info, dict):
            names = [channels_info.get('@Name', "Channel 0")]
        else:
            n_channels = self.get_code(scene_idx, 'C')
            names = [f"Channel {i}" for i in range(n_channels)]
        return names

    @unstable(doc=False)
    def view(self, scene: SceneIdx | None = None,
             channel: int = 0,
             depth: int | slice | np.ndarray | None = None,
             project_type: Literal['avg', 'max', 'min', 'std', 'median'] = 'max',
             norm: bool = True) -> np.ndarray:
        """
        Generates a view of the image data based on the provided parameters such as scene,
        channel, depth, projection type, and normalization. The function retrieves image data
        from either a mosaic file or a standard image file, processes it according to the
        specified depth, performs projection if required, and applies normalization if
        requested.

        :param scene: Scene index to be used for image loading. If None and the file is non-mosaic, defaults to 0.
              Used only for non-mosaic CZI files.
        :param channel: Index of the channel to be visualized. Defaults to 0.
        :param depth: Z-plane depth index, slice, or array. If None, all Z-slices are loaded.
              Can also be a specific integer index or an array of indices/slice.
        :param project_type: Type of Z-projection to apply if multiple Z-slices are selected.
              Options include 'avg', 'max', 'min', 'std', 'median'. Defaults to 'max'.
        :param norm: Whether to normalize the image intensity values. If True, divides the
              data by its maximum intensity value. Defaults to True.
        :return: A NumPy array containing the processed image data, which may be normalized
              and/or projected based on the input parameters.
        """

        kwargs = {}
        if self.is_mosaic:
            fn = self._czi_file.read_mosaic
            nz = np.max([self.get_code(i, 'Z') for i in range(self.n_scenes)])
        else:
            fn = self._czi_file.read_image
            scene = 0 if scene is None else scene
            nz = self.get_code(scene, 'Z')
            kwargs.update({'S': scene})

        #
        if depth is None:
            depth = np.arange(nz)
        if isinstance(depth, int):
            img = fn(C=channel, Z=depth, **kwargs).squeeze()
        elif isinstance(depth, (np.ndarray, slice)):
            stacks = np.array([fn(C=channel, Z=z, **kwargs)[0].squeeze() for z in depth])  # (C, )
            img = self.z_projection(stacks, project_type)
        else:
            raise TypeError(f'unknown type {type(depth)}')

        if norm:
            img = img / np.max(img)

        return img



import platform
import subprocess
from typing import TypedDict, Literal

import polars as pl

from neuralib.util.table import rich_data_frame_table
from neuralib.util.verbose import fprint

__all__ = [
    'print_gpu_table',
    'gpu_available',
    'check_mps_available',
    'check_nvidia_cuda_available'
]

RUN_BACKEND = Literal['tensorflow', 'torch']


def gpu_available(backend: RUN_BACKEND, *, check_smi: bool = False) -> bool:
    """
    Check if GPU is available.

    :param backend: {'torch', 'tensorflow'}
    :param check_smi: check if ``nvidia-smi`` is runnable
    :return:
    """
    system = platform.system()
    if system in ('win32', 'Windows', 'Linux'):
        return check_nvidia_cuda_available(backend=backend, check_smi=check_smi)
    elif system == 'Darwin':
        return check_mps_available(backend=backend)
    else:
        raise NotImplementedError(f'Unsupported system {system}')


def print_gpu_table(backend: RUN_BACKEND, *, check_smi: bool = False) -> None:
    """
    Print GPU info table and check the compatibility with backend package

    :param backend: {'torch', 'tensorflow'}
    :param check_smi: check if ``nvidia-smi`` is runnable
    :return:
    """
    system = platform.system()
    if system in ('win32', 'Windows', 'Linux'):
        info = _get_gpu_windows()
        check_nvidia_cuda_available(backend=backend, check_smi=check_smi)
    elif system == 'Darwin':
        info = _get_gpu_mac(backend=backend)
    else:
        raise NotImplementedError(f'{system}!')

    #
    if isinstance(info, dict):
        rich_data_frame_table(info)
    elif isinstance(info, list):
        df = pl.concat([pl.DataFrame(it) for it in info])
        rich_data_frame_table(df)
    else:
        raise TypeError('')


# ============= #
# Windows/Linux #
# ============= #

class GPUInfoWin(TypedDict, total=False):
    id: str
    name: str
    driver_version: str | None
    gpu_load: str
    """percentage of GPU usage"""
    total_memory: float
    """in MB"""
    free_memory: float
    used_memory: float
    temperature: float
    """in celsius"""


def _get_gpu_windows() -> list[GPUInfoWin]:
    import GPUtil

    ret = []
    gpus = GPUtil.getGPUs()
    for gpu in gpus:
        ret.append(
            GPUInfoWin(
                id=gpu.id,
                name=gpu.name,
                driver_version=gpu.driver,
                gpu_load=gpu.load,
                total_memory=gpu.memoryTotal,
                free_memory=gpu.memoryFree,
                used_memory=gpu.memoryUsed,
                temperature=gpu.temperature
            )
        )
    return ret


def check_nvidia_cuda_available(backend: RUN_BACKEND,
                                check_smi: bool = False) -> bool:
    """
    Checks if the GPU driver reacts and otherwise raises a custom error.
    Useful to check before long GPU-dependent processes.

    :param backend: {'torch', 'tensorflow'}
    :param check_smi: check if ``nvidia-smi`` is runnable
    """

    if check_smi:
        process = subprocess.Popen('nvidia-smi',
                                   shell=True,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        info, error = process.communicate()

        if process.returncode != 0:
            fprint(f"{error.decode('utf-8')}", vtype='warning')
        else:
            fprint("nvidia-smi command successful", vtype='pass')

    #
    is_available = False

    if backend == 'torch':
        import torch
        if torch.cuda.is_available():
            is_available = True
    elif backend == 'tensorflow':
        import tensorflow as tf
        if tf.test.is_built_with_cuda():
            is_available = True
    else:
        raise ValueError(f'unknown backend: {backend}')

    #
    if is_available:
        fprint(f'cuda is available in current env using backend {backend}', vtype='pass')
        return True
    else:
        return False


# ====== #
# Mac OS #
# ====== #


class GPUInfoMac(TypedDict, total=False):
    Chipset_Model: str
    Type: str
    Bus: str
    VRAM: str
    Vendor: str
    Device_ID: str
    Revision_ID: str
    Metal_Support: str
    mps_available: bool


def _get_gpu_mac(backend: RUN_BACKEND) -> GPUInfoMac:
    """get mac gpu info from subprocess

    :param backend: {'torch', 'tensorflow'}
    :return ``GPUInfoMac``
    """
    output = subprocess.check_output(["system_profiler", "SPDisplaysDataType"], universal_newlines=True)
    lines = output.splitlines()

    ret = {}
    cur_gpu = None

    for line in lines:
        line = line.strip()
        if len(line) == 0:
            continue
        elif line.startswith('Graphics/Displays:'):
            cur_gpu = {}
        elif line.startswith('Displays:'):
            break
        elif cur_gpu is not None and line.strip():
            key, value = line.split(':', 1)
            if value:
                ret[key.strip()] = value.strip()

    ret['mps_available'] = check_mps_available(backend=backend)

    return ret


def check_mps_available(backend: RUN_BACKEND) -> bool:
    """
    Check if metal is available

    :param backend: {'torch', 'tensorflow'}
    :return: bool
    """
    is_available = True

    if backend == 'torch':
        import torch
        if not torch.backends.mps.is_available():
            if not torch.backends.mps.is_built():
                fprint('MPS not available because pytorch install not built with MPS enable', vtype='warning')
            else:
                fprint('MPS not available because current MacOs version is not 12.3+,'
                       ' or do not have MPS-enabled device on this machine', vtype='warning')
            is_available = False

    elif backend == 'tensorflow':
        import tensorflow as tf
        if not tf.test.is_gpu_available():
            fprint('MPS not available in tensorflow backend', vtype='warning')
            is_available = False
    else:
        raise NotImplementedError(f'unknown backend: {backend}')

    #
    if is_available:
        fprint(f'MPS is available using backend: {backend}', vtype='pass')
        return True
    else:
        return False



import io
from pathlib import Path

import polars as pl

from neuralib.atlas.data import load_bg_structure_tree
from neuralib.io.core import ATLAS_CACHE_DIRECTORY
from neuralib.typing import PathLike
from neuralib.util.utils import ensure_dir
from neuralib.util.verbose import print_save

__all__ = ['load_cellatlas']


def load_cellatlas(file: PathLike | None = None, *,
                   with_cell_type: bool = False,
                   with_detail: bool = False,
                   with_total_neurons: bool = True,
                   with_acronym: bool = True,
                   reload: bool = False) -> pl.DataFrame:
    """
    Load the dataframe with cell types and volume information for each brain area

    .. seealso::

        `Rodarie D et al., (2022) <https://journals.plos.org/ploscompbiol/article?id=10.1371/journal.pcbi.1010739#sec047>`_

    :param file: Cached csv filepath. If not exist, download from the source paper
    :param with_cell_type: With cell type information, defaults to False
    :param with_detail:  With some outlier brain areas, defaults to False
    :param with_total_neurons: With ``n_neurons`` field, defaults to True
    :param with_acronym: With ``acronym`` field sync with structure tree data, defaults to True
    :param reload: Re-download the csv file
    :return: DataFrame
    """
    if file is None:
        file = ensure_dir(ATLAS_CACHE_DIRECTORY) / 'cellatlas.csv'

    if not Path(file).exists() or reload:
        df = _request(file).rename({'Brain region': 'name'})
    else:
        df = pl.read_csv(file).rename({'Brain region': 'name'})

    if not with_cell_type:
        df = df.select('name', 'Neuron [mm^-3]', 'Volumes [mm^3]')

    if not with_detail:
        patterns = (',', '/', r'\(')
        for pt in patterns:
            df = df.filter(~(pl.col('name').str.contains(pt)))

    if with_total_neurons:
        expr = (pl.col('Neuron [mm^-3]') * pl.col('Volumes [mm^3]')).alias('n_neurons').cast(pl.Int64)
        df = df.with_columns(expr).drop('Neuron [mm^-3]')

    if with_acronym:
        tree = load_bg_structure_tree().select('name', 'acronym').sort('name')
        df = df.join(tree, on='name')

    return df


def _request(output: Path) -> pl.DataFrame:
    """download from paper source"""
    import requests

    url = 'https://journals.plos.org/ploscompbiol/article/file?type=supplementary&id=10.1371/journal.pcbi.1010739.s011'
    resp = requests.get(url)

    if resp.status_code == 200:
        df = pl.read_excel(io.BytesIO(resp.content), sheet_name='Densities BBCAv1')
        df.write_csv(output)
        print_save(output, verb='DOWNLOAD')
    else:
        raise RuntimeError('download cellatlas FAIL')

    return df



from pathlib import Path
from typing import NamedTuple, Iterator

import matplotlib.pyplot as plt
import numpy as np
import vedo
from matplotlib.axes import Axes
from matplotlib.patches import Circle
from typing_extensions import Self, overload

from argclz import AbstractParser, argument
from neuralib.typing import PathLike

__all__ = [
    'SwcNode',
    'SwcFile',
    'plot_swc'
]

from argclz import pos_argument

Identifier = int
IdentifierName = str

IDENTIFIER_DICT: dict[Identifier, IdentifierName] = {
    0: 'undefined',
    1: 'soma',
    2: 'axon',
    3: 'basal',
    4: 'apical',
    5: 'custom'
}


class SwcNode(NamedTuple):
    n: int
    """node number"""

    identifier: Identifier
    """See IDENTIFIER_DICT"""

    x: float
    """position x"""
    y: float
    """position y"""
    z: float
    """position z"""
    r: float
    """radius"""
    parent: int
    """parent connectivity"""

    @property
    def identifier_name(self) -> IdentifierName:
        return IDENTIFIER_DICT.get(self.identifier, 'custom')

    @property
    def point(self) -> np.ndarray:
        return np.array([self.x, self.y, self.z])

    @property
    def is_undefined(self) -> bool:
        return self.identifier == 0

    @property
    def is_soma(self) -> bool:
        return self.identifier == 1

    @property
    def is_axon(self) -> bool:
        return self.identifier == 2

    @property
    def is_basal_dendrite(self) -> bool:
        return self.identifier == 3

    @property
    def is_apical_dendrite(self) -> bool:
        return self.identifier == 4

    @property
    def is_dendrite(self) -> bool:
        return self.is_basal_dendrite or self.is_apical_dendrite

    @property
    def is_custom(self) -> bool:
        return self.identifier >= 5


class SwcFile:
    """SWC File"""
    node: list[SwcNode]

    def __init__(self, node: list[SwcNode]):
        self.node = node

    @classmethod
    def load(cls, file: PathLike) -> Self:
        """
        :param file: swc filepath
        :return: ``SwcFile``
        """
        node = []
        with Path(file).open('r', encoding='Big5') as f:
            for line in f:
                line = line.strip()

                if len(line) == 0 or line.startswith('#'):
                    continue

                part = line.split()
                n = int(part[0])
                i = int(part[1])
                x = float(part[2])
                y = float(part[3])
                z = float(part[4])
                r = float(part[5])
                p = int(part[6])

                node.append(SwcNode(n, i, x, y, z, r, p))

        return cls(node)

    def __str__(self):
        line = [str(node) for node in self.node]
        return '\n'.join(line)

    @overload
    def __getitem__(self, item: int) -> SwcNode:
        pass

    @overload
    def __getitem__(self, item: IdentifierName) -> Self:
        pass

    def __getitem__(self, item: int | str) -> SwcNode | Self:
        if isinstance(item, int):
            try:
                ret = self.node[item - 1]  # to index
            except IndexError:
                ret = None

            if ret is not None and ret.n == item:
                return ret

            raise KeyError(f'item not found: {item}, might be loss parent connection')

        elif isinstance(item, str):
            if item == 'soma':
                node = [n for n in self.foreach_node() if n.is_soma]
            elif item == 'axon':
                node = [n for n in self.foreach_node() if n.is_axon]
            elif item == 'dendrite':
                node = [n for n in self.foreach_node() if n.is_dendrite]
            elif item == 'basal':
                node = [n for n in self.foreach_node() if n.is_basal_dendrite]
            elif item == 'apical':
                node = [n for n in self.foreach_node() if n.is_apical_dendrite]
            elif item == 'dendrite':
                node = [n for n in self.foreach_node() if n.is_dendrite]
            elif item == 'custom':
                node = [n for n in self.foreach_node() if n.is_custom]
            elif item == 'undefined':
                node = [n for n in self.foreach_node() if n.is_undefined]
            else:
                raise ValueError('')

            return SwcFile(node)

        else:
            raise TypeError(f'item must be int or str: {type(item)}')

    @property
    def points(self) -> np.ndarray:
        return np.array([[n.x, n.y, n.z] for n in self.foreach_node()])

    @property
    def radii(self) -> np.ndarray:
        return np.array([n.r for n in self.foreach_node()])

    @property
    def parents(self) -> np.ndarray:
        return np.array([n.parent for n in self.foreach_node()])

    @property
    def unique_identifier(self) -> list[IdentifierName]:
        idfs = np.unique([n.identifier for n in self.foreach_node()])
        return [
            IDENTIFIER_DICT.get(idf, 'custom')
            for idf in idfs
        ]

    def foreach_identifier(self, as_dict: bool) -> list[Self] | dict[str, Self]:
        if as_dict:
            return {idf: self[idf] for idf in self.unique_identifier}
        else:
            return [self[idf] for idf in self.unique_identifier]

    def foreach_node(self) -> Iterator[SwcNode]:
        for node in self.node:
            yield node

    def foreach_line(self) -> Iterator[tuple[SwcNode, SwcNode]]:
        for node in self.node:
            if node.parent > 0:
                yield node, self[node.parent]


# ============== #
# Plot Functions #
# ============== #

Point3D = tuple[float, float, float]
Point2D = tuple[float, float]

DEFAULT_COLOR: dict[IdentifierName, str] = {
    'soma': 'b',
    'axon': 'r',
    'dendrite': 'k',
    'undefined': 'k',
    'custom': 'k'
}


def plot_swc(swc: SwcFile,
             radius: bool = True,
             color: dict[str, str] | None = None,
             as_2d: bool = False):
    """
    Plot swc file as 2d

    :param swc: ``SwcFile``
    :param radius: Plot with radius.
    :param color: Color dict. With {identifier name: color coded}
    :param as_2d:
    """
    if color is None:
        color = DEFAULT_COLOR

    if as_2d:
        _plot_swc_2d(swc, radius, color)
    else:
        _plot_swc_3d(swc, radius, color)


def projection_2d(p: Point3D) -> Point2D:
    """Default projection function, remove z value.

    :param p: 3d points
    :return: 2d points
    """
    return p[0], p[1]


def smooth_line_radius(ax: Axes,
                       p1: Point2D,
                       p2: Point2D,
                       r1: float,
                       r2: float,
                       num: int = 2,
                       **kwargs):
    """

    :param ax: ``Axes``
    :param p1: Point 1
    :param p2: Point 2
    :param r1: Radius 1
    :param r2: Radius 2
    :param num: Number of segments
    :param kwargs: Additional arguments pass to ``plt.plot()``
    :return:
    """
    px = np.linspace(p1[0], p2[0], num + 1)
    py = np.linspace(p1[1], p2[1], num + 1)
    lw = np.linspace(r1, r2, num)
    for i in range(num):
        ax.plot(px[i:i + 2], py[i:i + 2], lw=lw[i], **kwargs)


def _plot_swc_2d(swc, radius, color):
    fig, ax = plt.subplots()
    for n1, n2 in swc.foreach_line():
        c = color.get(n1.identifier_name, 'k')

        p1 = projection_2d((n1.x, n1.y, n1.z))
        p2 = projection_2d((n2.x, n2.y, n2.z))

        if radius:
            if n2.is_soma:
                ax.add_artist(Circle(p2, n2.r, color=color['soma']))
                if not n1.is_soma:
                    smooth_line_radius(ax, p1, p2, n1.r, n1.r, color=c, solid_capstyle='round')
            else:
                smooth_line_radius(ax, p1, p2, n1.r, n2.r, color=c, solid_capstyle='round')
        else:
            px = p1[0], p2[0]
            py = p1[1], p2[1]
            ax.plot(px, py, color=c, solid_capstyle='round')

    ax.axis('off')
    ax.set_xticklabels([])
    ax.set_yticklabels([])
    plt.show()


def _plot_swc_3d(swc: SwcFile,
                 radius,
                 color,
                 spheres_size: float = 3,
                 lw: float = 5):
    plotter = vedo.Plotter()

    axons = []
    axons_line = []
    axons_radii = []

    dendrites = []
    dendrites_line = []
    dendrites_radii = []

    somata = []
    somata_line = []
    somata_radii = []

    other = []
    other_line = []
    other_radii = []

    for i, n in enumerate(swc.foreach_node()):
        if n.parent == -1:
            if n.is_soma:
                somata.append([n.x, n.y, n.z])
                somata_radii.append(10)
            continue

        r = n.r * spheres_size if radius else 5

        if n.is_axon:
            axons.append([n.x, n.y, n.z])
            axons_line.append([n.parent - 1, i])  # Use parent-child connection for axons
            axons_radii.append(r)
        elif n.is_dendrite:
            dendrites.append([n.x, n.y, n.z])
            dendrites_line.append([n.parent - 1, i])
            dendrites_radii.append(r)
        elif n.is_soma:
            somata.append([n.x, n.y, n.z])
            somata_line.append([n.parent - 1, i])
            somata_radii.append(10)  # fix value
        elif n.is_undefined or n.is_custom:
            other.append([n.x, n.y, n.z])
            other_line.append([n.parent - 1, i])
            other_radii.append(r)

    #
    if 'soma' in swc.unique_identifier:
        soma_spheres = vedo.Spheres(somata, r=somata_radii, c=color['soma'])
        soma_lines = vedo.Lines(swc.points[somata_line], c=color['soma'], lw=lw)
        plotter += soma_spheres
        plotter += soma_lines

    if 'dendrite' in swc.unique_identifier:
        dendrite_spheres = vedo.Spheres(dendrites, r=dendrites_radii, c=color['dendrite'])
        dendrite_lines = vedo.Lines(swc.points[dendrites_line], c=color['dendrite'], lw=lw)
        plotter += dendrite_spheres
        plotter += dendrite_lines

    if 'axon' in swc.unique_identifier:
        axon_spheres = vedo.Spheres(axons, r=axons_radii, c=color['axon'])
        axon_lines = vedo.Lines(swc.points[axons_line], c=color['axon'], lw=lw)
        plotter += axon_spheres
        plotter += axon_lines

    if len(other) > 0:
        other_spheres = vedo.Spheres(other, r=other_radii, c=color.get('custom', 'k'))
        other_lines = vedo.Lines(swc.points[other_line], c=color.get('custom', 'k'), lw=lw)
        plotter += other_spheres
        plotter += other_lines

    plotter.show()


# ======== #
# Plot CLI #
# ======== #

class SwcPlotOptions(AbstractParser):
    file: str = pos_argument(
        'FILE',
        help='filepath of the swc file'
    )

    radius: bool = argument(
        '-R', '--radius',
        help='Whether plot with radius'
    )

    as_2d: bool = argument(
        '--2d',
        help='Whether plot with 2d, otherwise, plot as 3d'
    )

    def run(self):
        swc = SwcFile.load(self.file)
        plot_swc(swc, radius=self.radius, as_2d=self.as_2d)


if __name__ == '__main__':
    SwcPlotOptions().main()



from __future__ import annotations

import abc
import math
import warnings
from typing import Final, ClassVar, Literal, get_args

import attrs
import matplotlib.pyplot as plt
import numpy as np
from brainglobe_atlasapi import BrainGlobeAtlas
from matplotlib.axes import Axes
from matplotlib.colors import ListedColormap
from matplotlib.patches import Patch
from typing_extensions import Self

from neuralib.atlas.data import ATLAS_NAME
from neuralib.atlas.typing import PLANE_TYPE
from neuralib.atlas.util import ALLEN_CCF_10um_BREGMA

__all__ = [
    'VIEW_TYPE',
    'get_slice_view',
    'AbstractSliceView',
    'SlicePlane'
]

VIEW_TYPE = Literal['annotation', 'reference']
"""View type for the slice"""


def get_slice_view(view: VIEW_TYPE,
                   plane_type: PLANE_TYPE, *,
                   name: str = 'allen_mouse',
                   resolution: int = 10,
                   check_latest: bool = True) -> AbstractSliceView:
    """
    Load the mouse brain slice view

    .. seealso::

        `brainglobe-atlasap doc <https://brainglobe.info/documentation/brainglobe-atlasapi/usage/atlas-details.html>`_

    :param view: ``VIEW_TYPE``.
    :param plane_type: ``PLANE_TYPE``. {'coronal', 'sagittal', 'transverse'}
    :param name: Name of the atlas.
    :param resolution: Volume resolution in um. default is 10 um
    :param check_latest: If True, check the latest version of brain
    :return: :class:`AbstractSliceView`
    """
    match view:
        case 'annotation' | 'reference':
            atlas_name = f'{name}_{resolution}um'
            if atlas_name not in get_args(ATLAS_NAME):
                raise ValueError(f'{atlas_name} not found or not implemented')
            data = getattr(BrainGlobeAtlas(atlas_name, check_latest=check_latest), view)
        case _:
            raise ValueError(f'Unknown view: {view}')

    return AbstractSliceView(view, plane_type, resolution, data)


class AbstractSliceView(metaclass=abc.ABCMeta):
    """
    SliceView ABC for different `plane type`

    `Dimension parameters`:

        AP = anterior-posterior

        DV = dorsal-ventral

        ML = medial-lateral

        W = view width

        H = view height
    """
    REFERENCE_FROM: ClassVar[str] = ''
    """reference from which axis"""

    view_type: Final[VIEW_TYPE]
    """``VIEW_TYPE``. {'annotation', 'reference'}"""

    plane_type: Final[PLANE_TYPE]
    """`PLANE_TYPE``. {'coronal', 'sagittal', 'transverse'}"""

    resolution: Final[int]
    """um/pixel"""

    reference: Final[np.ndarray]
    """Array[float, [AP, DV, ML]]"""

    grid_x: Final[np.ndarray]
    """Array[int, [W, H]]"""

    grid_y: Final[np.ndarray]
    """Array[int, [W, H]]"""

    def __new__(cls, view_type: VIEW_TYPE,
                plane: PLANE_TYPE,
                resolution: int,
                reference: np.ndarray):
        if plane == 'coronal':
            return object.__new__(CoronalSliceView)
        elif plane == 'sagittal':
            return object.__new__(SagittalSliceView)
        elif plane == 'transverse':
            return object.__new__(TransverseSliceView)
        else:
            raise ValueError(f'invalid plane: {plane}')

    def __init__(self, view_type: VIEW_TYPE,
                 plane: PLANE_TYPE,
                 resolution: int,
                 reference: np.ndarray):
        """

        :param view_type: ``DATA_SOURCE_TYPE``. {'ccf_annotation', 'ccf_template', 'allensdk_annotation'}
        :param plane: `PLANE_TYPE``. {'coronal', 'sagittal', 'transverse'}
        :param resolution: um/pixel
        :param reference: Array[uint16, [AP, DV, ML]]
        """
        self.view_type = view_type
        self.plane_type = plane
        self.resolution = resolution
        self.reference = reference

        self.grid_y, self.grid_x = np.mgrid[0:self.height, 0:self.width]

        self._check_attrs()

    def _check_attrs(self):
        if self.resolution == 10:
            assert self.reference.shape == (1320, 800, 1140)
        elif self.resolution == 25:
            assert self.reference.shape == (528, 320, 456)

    @property
    def bregma(self) -> np.ndarray:
        if self.resolution == 10:
            return ALLEN_CCF_10um_BREGMA
        raise NotImplementedError('')

    @property
    def n_ap(self) -> int:
        """number of slices along AP axis"""
        return self.reference.shape[0]

    @property
    def n_dv(self) -> int:
        """number of slices along DV axis"""
        return self.reference.shape[1]

    @property
    def n_ml(self) -> int:
        """number of slices along ML axis"""
        return self.reference.shape[2]

    @property
    @abc.abstractmethod
    def n_planes(self) -> int:
        """number of planes in a specific plane view"""
        pass

    @property
    @abc.abstractmethod
    def width(self) -> int:
        """width (pixel) in a specific plane view"""
        pass

    @property
    @abc.abstractmethod
    def height(self) -> int:
        """height (pixel) in a specific plane view"""
        pass

    @property
    def width_mm(self) -> float:
        """width (um) in a specific plane view"""
        return self.width * self.resolution / 1000

    @property
    def height_mm(self) -> float:
        """height (um) in a specific plane view"""
        return self.height * self.resolution / 1000

    @property
    @abc.abstractmethod
    def reference_point(self) -> int:
        """reference point in a specific plane view. aka, bregma plane index"""
        pass

    @property
    @abc.abstractmethod
    def project_index(self) -> tuple[int, int, int]:
        """plane(p), x, y of index order in (AP, DV, ML)

        :return: (p, x, y)
        """
        pass

    @property
    @abc.abstractmethod
    def max_projection_axis(self) -> tuple[int, int, int]:
        pass

    def plot_max_projection(self, ax: Axes, *,
                            annotation_regions: str | list[str] | None = None,
                            annotation_cmap: str = 'hsv'):
        """
        Plot max projection for the given ``plane_type``

        :param ax: ``Axes``
        :param annotation_regions: annotation_regions
        :param annotation_cmap: camp for the annotation regions, defaults to 'hsv'
        """

        img = self.reference.max(axis=self.max_projection_axis)
        if isinstance(self, SagittalSliceView):
            img = img.T

        ext = _get_xy_range(self, to_um=True)
        ax.imshow(img, cmap='Greys', extent=ext)

        if annotation_regions is not None:
            from neuralib.atlas.data import get_leaf_in_annotation
            annotation = get_slice_view('annotation', self.plane_type, resolution=self.resolution).reference

            if isinstance(annotation_regions, str):
                annotation_regions = [annotation_regions]

            #
            region_colors = plt.get_cmap(annotation_cmap, len(annotation_regions))
            for i, r in enumerate(annotation_regions):
                ids = get_leaf_in_annotation(r, name=False)
                mask = np.isin(annotation, ids)

                region_mask = np.full(annotation.shape, np.nan, dtype=np.float64)
                region_mask[mask] = 1.0

                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", category=RuntimeWarning)
                    proj = np.nanmax(region_mask, axis=self.max_projection_axis)

                if isinstance(self, SagittalSliceView):
                    proj = proj.T

                masked = np.ma.masked_invalid(proj)
                ax.imshow(
                    masked,
                    cmap=ListedColormap(region_colors(i)),
                    extent=ext,
                    alpha=0.7,
                    zorder=2 + i,
                )

            #
            ax.set(xlabel='um', ylabel='um')
            legend_elements = [
                Patch(facecolor=region_colors(i), label=region, alpha=0.7)
                for i, region in enumerate(annotation_regions)
            ]
            ax.legend(handles=legend_elements, title="Regions", loc='upper right')

    def plane_at(self, slice_index: int) -> 'SlicePlane':
        return SlicePlane(slice_index, int(self.width // 2), int(self.height // 2), 0, 0, self)

    def offset(self, h: int, v: int) -> np.ndarray:
        """

        :param h: horizontal plane diff to the center. right side positive.
        :param v: vertical plane diff to the center. bottom side positive.
        :return: (H, W) array
        """
        x_frame = np.round(np.linspace(-h, h, self.width)).astype(int)
        y_frame = np.round(np.linspace(-v, v, self.height)).astype(int)
        return np.add.outer(y_frame, x_frame)

    def plane(self, offset: int | tuple[int, int, int] | np.ndarray) -> np.ndarray:
        """Get image plane.

        :param offset: Array[int, height, width] or tuple (plane, dh, dv)
        :return:
        """
        if isinstance(offset, int):
            offset = np.full_like((self.height, self.width), offset)
        elif isinstance(offset, tuple):
            offset = offset[0] + self.offset(offset[1], offset[2])
        elif not isinstance(offset, np.ndarray):
            raise TypeError(str(type(offset)))

        offset[offset < 0] = 0
        offset[offset > self.n_planes] = self.n_planes - 1

        return self.reference[self.coor_on(offset, (self.grid_x, self.grid_y))]

    def coor_on(self, plane: np.ndarray,
                o: tuple[np.ndarray, np.ndarray]) -> tuple[np.ndarray, ...]:
        """
        map slice point (x, y) at plane *plane* back to volume point (ap, dv, ml)

        :param plane: plane number of array
        :param o: tuple of (x, y)
        :return: (ap, dv, ml)
        """
        pidx, xidx, yidx = self.project_index
        ret = [0, 0, 0]
        ret[pidx] = plane
        ret[xidx] = o[0]
        ret[yidx] = o[1]

        return tuple(ret)


class CoronalSliceView(AbstractSliceView):
    REFERENCE_FROM: ClassVar[str] = 'AP'

    @property
    def n_planes(self) -> int:
        return self.n_ap

    @property
    def width(self) -> int:
        return self.n_ml

    @property
    def height(self) -> int:
        return self.n_dv

    @property
    def reference_point(self) -> int:
        return int(self.bregma[0])

    @property
    def project_index(self) -> tuple[int, int, int]:
        return 0, 2, 1

    @property
    def max_projection_axis(self) -> int:
        return 0


class SagittalSliceView(AbstractSliceView):
    REFERENCE_FROM: ClassVar[str] = 'ML'

    @property
    def n_planes(self) -> int:
        return self.n_ml

    @property
    def width(self) -> int:
        return self.n_ap

    @property
    def height(self) -> int:
        return self.n_dv

    @property
    def reference_point(self) -> int:
        return int(self.bregma[2])

    @property
    def project_index(self) -> tuple[int, int, int]:
        return 2, 0, 1  # p=ML, x=AP, y=DV

    @property
    def max_projection_axis(self) -> int:
        return 2


class TransverseSliceView(AbstractSliceView):
    REFERENCE_FROM: ClassVar[str] = 'DV'

    @property
    def n_planes(self) -> int:
        return self.n_dv

    @property
    def width(self) -> int:
        return self.n_ml

    @property
    def height(self) -> int:
        return self.n_ap

    @property
    def reference_point(self) -> int:
        return int(self.bregma[1])

    @property
    def project_index(self) -> tuple[int, int, int]:
        return 1, 2, 0

    @property
    def max_projection_axis(self):
        return 1


@attrs.define
class SlicePlane:
    """2D Wrapper for a specific plane"""

    slice_index: int
    """anchor index"""

    ax: int
    """anchor x"""

    ay: int
    """anchor y"""

    dw: int
    """dw in um"""

    dh: int
    """dh in um"""

    slice_view: AbstractSliceView
    """``AbstractSliceView``"""

    unit: str = 'a.u.'

    @property
    def image(self) -> np.ndarray:
        return self.slice_view.plane(self.plane_offset)

    @property
    def plane_offset(self) -> np.ndarray:
        offset = self.slice_view.offset(self.dw, self.dh)
        return self.slice_index + offset - offset[self.ay, self.ax]

    @property
    def reference_value(self) -> float:
        """relative to reference point"""
        factor = 1000 / self.slice_view.resolution
        return round((self.slice_view.reference_point - self.slice_index) / factor, 2)

    def with_offset(self, dw: int, dh: int, debug: bool = False) -> Self:
        if debug:
            deg_x, deg_y = self._value_to_angle(dw, dh)
            print(f'{dw=}, {dh=}')
            print(f'{deg_x=}, {deg_y=}')

        return attrs.evolve(self, dw=int(dw), dh=int(dh))

    def _value_to_angle(self, dw: int, dh: int) -> tuple[float, float]:
        """delta value to degree"""
        rx = math.atan(2 * dw / self.slice_view.width)
        ry = math.atan(2 * dh / self.slice_view.height)
        deg_x = np.rad2deg(rx)
        deg_y = np.rad2deg(ry)

        return deg_x, deg_y

    def with_angle_offset(self, deg_x: float = 0, deg_y: float = 0) -> Self:
        """
        with degree offset

        :param deg_x: degree in x axis (width)
        :param deg_y: degree in y axis (height)
        :return:
        """
        rx = np.deg2rad(deg_x)
        ry = np.deg2rad(deg_y)

        dw = int(self.slice_view.width * math.tan(rx) / 2)
        dh = int(self.slice_view.height * math.tan(ry) / 2)

        return self.with_offset(dw, dh)

    def plot(self,
             ax: Axes | None = None,
             to_um: bool = True,
             annotation_region: str | list[str] | None = None,
             boundaries: bool = False,
             with_title: bool = False,
             extent: tuple[float, float, float, float] | None = None,
             reference_bg_value: float | None = None,
             annotation_cmap: str = 'berlin',
             annotation_rescale: bool = True,
             **kwargs) -> None:
        """
        :param ax: The Axes object on which to plot. If None, a new figure and axes are created.
        :param to_um: A boolean flag indicating whether the coordinates should be converted to micrometers. Defaults to True.
            Only applicable if ``extent`` is None.
        :param annotation_region: The annotation region on which to plot. Defaults to None.
        :param boundaries: A boolean indicating whether to include annotations in the plot.
        :param with_title: A boolean indicating whether to include a title in the plot.
        :param extent: A tuple defining the image boundaries (left, right, bottom, top). If None, boundaries are computed internally.
        :param reference_bg_value: If specified, remove background of its value in when view_type is **reference** (i.e., set as 10).
        :param annotation_cmap: Cmap for the annotation regions if specified
        :param annotation_rescale: Rescale the image when view_type is **annotation**.
        :param kwargs: Additional keyword arguments passed to ``ax.imshow()``.
        """

        if ax is None:
            _, ax = plt.subplots()

        if extent is None:
            extent = self._get_xy_range(to_um)

        # value modify
        image = self.image.astype(float)
        if reference_bg_value is not None and self.slice_view.view_type == 'reference':
            image[image < reference_bg_value] = np.nan

        if annotation_rescale and self.slice_view.view_type == 'annotation':
            image[image == 0] = np.nan
            valid = ~np.isnan(image)
            unique_vals = np.unique(image[valid])
            remapped_image = np.full_like(image, np.nan)
            remapped_image[valid] = np.searchsorted(unique_vals, image[valid]) + 1
            image = remapped_image

        # image
        ax.imshow(image, cmap='Greys', extent=extent, clip_on=False, **kwargs)

        # annotation region
        if annotation_region is not None:
            self.plot_annotation_regions(ax, annotation_region, extent, annotation_cmap, **kwargs)

        # with boundaries
        if boundaries:
            self.plot_boundaries(ax, extent, **kwargs)

        #
        if with_title:
            ax.set_title(f'{self.slice_view.REFERENCE_FROM}: {self.reference_value} mm')

        ax.set(xlabel=self.unit, ylabel=self.unit)

    def plot_annotation_regions(self, ax, regions, extent, cmap='berlin', **kwargs):
        from matplotlib.patches import Patch
        from neuralib.atlas.data import get_leaf_in_annotation

        annotation = (
            get_slice_view('annotation', self.slice_view.plane_type, resolution=self.slice_view.resolution)
            .plane(self.plane_offset)
            .astype(float)
        )

        if isinstance(regions, str):
            regions = [regions]

        area = np.full_like(annotation, np.nan)
        for i, r in enumerate(regions):
            ids = get_leaf_in_annotation(r, name=False)
            mx = np.isin(annotation, ids)
            area[mx] = i + 1.0

        cmap = plt.get_cmap(cmap, len(regions))

        ax.imshow(area, cmap=cmap, extent=extent, alpha=0.9, clip_on=False, **kwargs)
        legend_elements = [
            Patch(facecolor=cmap(i / len(regions)), label=region)
            for i, region in enumerate(regions)
        ]
        ax.legend(handles=legend_elements, title="Regions", loc='upper right')

    def plot_boundaries(self, ax, extent, cmap='binary', alpha=0.5, **kwargs):
        """
        Plot the annotation boundaries

        :param ax: ``Axes``
        :param extent: A tuple defining the image boundaries (left, right, bottom, top). If None, boundaries are computed internally.
        :param cmap: Colormap to be used for the annotation image. Defaults to 'binary'.
        :param alpha: The imshow alpha, between 0 (transparent) and 1 (opaque). Defaults to 0.5.
        """
        from skimage.segmentation import find_boundaries

        ann_img = (
            get_slice_view('annotation', self.slice_view.plane_type, resolution=self.slice_view.resolution)
            .plane(self.plane_offset)
        )

        ann = find_boundaries(ann_img, mode='outer').astype(float)
        ann[ann == 0] = np.nan

        ax.imshow(ann, cmap=cmap, extent=extent, alpha=alpha,
                  clip_on=False, interpolation='none', vmin=0, vmax=1, **kwargs)

    def _get_xy_range(self, to_um: bool = True) -> tuple[float, float, float, float]:
        self.unit = 'um' if to_um else 'mm'
        return _get_xy_range(self.slice_view, to_um=to_um)


def _get_xy_range(view: AbstractSliceView, to_um: bool = True) -> tuple[float, float, float, float]:
    if to_um:
        x0 = -view.width_mm / 2 * 1000
        x1 = view.width_mm / 2 * 1000
        y0 = view.height_mm * 1000
        y1 = 0
    else:
        x0 = view.width_mm / 2
        x1 = view.width_mm / 2
        y0 = view.height
        y1 = 0

    return x0, x1, y0, y1


import argparse

from neuralib.dashboard import BokehServer
from .view_all import AllView
from .view_animal import AnimalView
from .view_figure import AnimalFigureView

ap = argparse.ArgumentParser()
ap.add_argument('--animal', metavar='NAME', default=None)
opt = ap.parse_args()

open_url = '/'
if opt.animal is not None:
    open_url = f'/pool?animal={opt.animal}'

BokehServer().start({
    '/': AllView(),
    '/pool': AnimalView(),
    '/pool/figure': AnimalFigureView(),
}, open_url=open_url)



from collections.abc import Callable
from typing import Literal, Union, Iterable

import numpy as np
from neuralib.typing import ArrayLike

__all__ = [
    'SegmentLike',
    'segment_bool_mask',
    'is_sorted',
    'segment_mask',
    'segment_epochs',
    'has_gap',
    'segment_gap',
    'as_segment',
    'segment_range',
    'segment_duration',
    'segment_at_least_duration',
    'segment_expand_duration',
    'segment_universe',
    'segment_flatten',
    'segment_invert',
    'segment_intersection',
    'segment_union',
    'segment_diff',
    'segment_contains',
    'segment_index',
    'segment_overlap',
    'segment_overlap_index',
    'segment_join',
    'segment_join_index',
    'segment_map',
    'segment_group_map',
    'segment_sample',
    'segment_bins',
    'shuffle_time',
    'shuffle_time_uniform',
    'shuffle_time_normal',
    'shift_time',
    'foreach_map',
    'grouped_iter'
]

Segment = np.ndarray  # (N, 2) value array ([(start, stop)]), as a segment.
SegmentLike = Union[Segment, tuple[float, float], list[tuple[float, float]]]
SegmentGroup = np.ndarray  # (N,) int array, where a unique value indicate a unique group/segment.


def segment_bool_mask(mx: np.ndarray) -> Segment:
    """
    :param mx: `Array[bool, N]`
    :return: ``Segment``. `Array[int, [N, 2]]`
    """
    if mx.ndim != 1:
        raise ValueError('not a (N, 1) segment')

    if mx.dtype != np.bool_:
        raise TypeError('not a bool array')

    diff = np.diff(mx.astype(int), prepend=0, append=0)
    starts = np.where(diff == 1)[0]
    ends = np.where(diff == -1)[0]

    return np.column_stack((starts, ends))


def is_sorted(a: np.ndarray, strict: bool = False) -> bool:
    """
    Check if array is sorted.
    [reference](https://stackoverflow.com/a/47004507)

    :param a: Input array to check for sorted order.
    :param strict: If True, checks for strictly increasing order. Otherwise, checks for non-decreasing order.
    :return: Returns True if the input array is sorted based on the specified criteria, else False.
    """
    if strict:
        return np.all(a[:-1] < a[1:])
    else:
        return np.all(a[:-1] <= a[1:])


def segment_mask(x: np.ndarray,
                 t: np.ndarray | None = None,
                 duration: float | None = None,
                 merge: float | None = None) -> SegmentGroup:
    """segmenting an array that the value is long enough.

    :param x: (N,) int-cast-able array.
    :param t: (N,) T-value array
    :param duration: T. the minimal True duration, duration smaller than value will set to even value.
        If using negative value, then duration larger than value will set to even value.
    :param merge: T. the minima False duration, duration smaller than value will set to odd value.
    :return: (N,) int array, where even value indicate False, odd value indicate True
    """
    if x.ndim != 1:
        raise ValueError('x ndim not 1')
    if t is None:
        t = np.arange(len(x))
    elif not is_sorted(t, strict=True):
        raise ValueError('t not sorted')

    if x.shape != t.shape:
        raise ValueError('t.shape != x.shape')
    if len(x) == 0:
        return np.array([], dtype=int)

    edge = np.sign(np.diff(x.astype(int), prepend=0))
    s = np.cumsum(np.abs(edge))
    sx = s[-1]

    if merge is not None:
        for i in range(2, sx + 1, 2):
            if len(d := t[s == i]) > 0:
                if d[-1] - d[0] <= merge:
                    s[s == i - 1] = i + 1
                    s[s == i] = i + 1

    if duration is not None:
        if duration > 0:
            for i in range(1, sx + 1, 2):
                if len(d := t[s == i]) > 0:
                    if d[-1] - d[0] <= duration:
                        s[s == i] = i - 1
        elif duration < 0:
            for i in range(1, sx + 1, 2):
                if len(d := t[s == i]) > 0:
                    if d[-1] - d[0] >= duration:
                        s[s == i] = i - 1
    return s


def segment_epochs(x: np.ndarray,
                   t: np.ndarray | None = None,
                   duration: float | None = None,
                   merge: float | None = None) -> Segment:
    """

    :param x: (N,) int-cast-able array.
    :param t: (N,) T-value array.
    :param duration:
    :param merge:
    :return: (M, 2) T-value array. If *t* is None, return an N-value index array
    """
    if t is None:
        t = np.arange(len(x))

    if x.shape != t.shape:
        raise ValueError('t.shape != x.shape')

    if len(x) == 0:
        return np.zeros((0, 2), dtype=t.dtype)

    g = segment_mask(x, t, duration, merge)

    ret = []
    for gg in range(1, np.max(g) + 1, 2):
        if len(i := np.nonzero(g == gg)[0]):
            if t is None:
                ret.append((i[0], i[-1]))
            else:
                ret.append((t[i[0]], t[i[-1]]))
    return np.array(ret)


def has_gap(y: np.ndarray, gap: float) -> bool:
    """

    :param y: (N,) V-value array
    :param gap: V
    :return:
    """
    return np.any(np.abs(np.diff(y)) > gap)


def segment_gap(x: np.ndarray, gap: float) -> SegmentGroup:
    """segmenting an array that cut at the place which the difference of nearby value larger than *gap*.

    `min{|ai-aj|} > gap` for any value `ai` from segment `si`, any value `aj` from segment `sj`, `si != sj`.

    :param x: (N,) V-value array
    :param gap: V
    :return: (N,) int-group array
    """
    if len(x) == 0:
        raise ValueError('empty array')
    elif len(x) == 1:
        return np.array([0])
    else:
        return np.cumsum(np.abs(np.diff(x, prepend=x[0])) > gap)


def as_segment(segs: SegmentLike) -> Segment:
    """
    :param segs: Input segment-like data that can be converted to a 2D array of floats.
    :return: A numpy 2D array representation of the input segments.
    :raises ValueError: If the input cannot be reshaped to a (N, 2) segment array.
    """
    ret = np.atleast_2d(segs).astype(float, copy=False)
    if ret.ndim != 2 or ret.shape[1] != 2:
        raise ValueError(f'not a (N, 2) segment array : {ret.shape}')
    return ret


def segment_range(segs: SegmentLike) -> np.ndarray:
    """change [(start, stop)] to [(start, duration)]

    :param segs: segments. `Array[float, [N, 2]]`
    :return: `Array[int, [N, 2]]`
    """
    a = as_segment(segs)
    r = a.copy()
    r[:, 1] = a[:, 1] - a[:, 0]
    return r


def segment_duration(segs: SegmentLike) -> np.ndarray:
    """

    :param segs: T-value segments. `Array[float, [N, 2]]`
    :return: T-value array. `Array[int, N]`
    """
    a = as_segment(segs)
    return a[:, 1] - a[:, 0]


def segment_at_least_duration(segs: SegmentLike, duration: float) -> np.ndarray:
    """

    :param segs: T-value segments. `Array[float, [N, 2]]`
    :param duration: T-value
    :return: T-value segments. `Array[float, [N, 2]]`
    """
    a = as_segment(segs).copy()
    dur = duration - (a[:, 1] - a[:, 0])
    ext = np.where(dur > 0, dur / 2, 0)
    a[:, 0] -= ext
    a[:, 1] += ext

    return a


def segment_expand_duration(segs: SegmentLike, duration: float | tuple[float, float] | np.ndarray) -> np.ndarray:
    """

    :param segs: T-value segments. `Array[float, [N, 2]]`
    :param duration: T-value scalar, tuple(prepend:T, append:T), or (N, 2?) array
    :return: T-value segments. `Array[float, [N, 2]]`
    """
    a = as_segment(segs).copy()

    if isinstance(duration, tuple):
        if len(duration) != 2:
            raise ValueError()

        ext = duration

    elif isinstance(duration, np.ndarray):
        if duration.ndim == 1:
            ext = duration / 2
            ext = ext, ext
        elif duration.ndim == 2:
            ext = duration
        else:
            raise ValueError()

    else:
        ext = duration / 2, duration / 2

    a[:, 0] -= ext[0]
    a[:, 1] += ext[1]

    return a


def segment_universe() -> Segment:
    """

    :return: segment of [-inf, inf]
    """
    return np.array([[-np.inf, np.inf]])


def segment_flatten(segs: SegmentLike, closed: bool = True) -> Segment:
    """sorting and remove overlapped segments.

    :param segs: T-value segments. `Array[float, [N, 2]]`
    :param closed: Is segment a closed on right side?
    :return: T-value segments. `Array[float, [M, 2]]`
    """
    a = as_segment(segs)

    if len(a) == 0:
        return a

    if not is_sorted(a[:, 0]):
        a = a[np.argsort(a[:, 0])]

    if is_sorted(a.ravel(), strict=True):
        return a

    r = np.empty_like(a)
    r[0] = a[0]
    i = 1
    j = 1
    while j < len(a):
        p: float = r[i - 1, 1]
        q: float = a[j, 0]
        assert r[i - 1, 0] <= q, f'{r[i-1]=} <> {a[j]=}'

        if (p >= q if closed else p > q):
            r[i - 1, 1] = max(a[j, 1], p)
        else:
            r[i] = a[j]
            i += 1
        j += 1

    r = r[:i]
    return r


def segment_invert(segs: SegmentLike) -> Segment:
    """

    :param segs: T-value segments. `Array[float, [N, 2]]`
    :return:  T-value segments. `Array[float, [N+1, 2]]`
    """
    return _segment_invert(segment_flatten(segs, closed=False))


def _segment_invert(segs: Segment) -> Segment:
    if len(segs) == 0:
        return segment_universe()

    a = np.concatenate([[-np.inf], segs.ravel(), [np.inf]]).reshape((-1, 2))

    d = []
    if np.all(np.isinf(a[0])):
        d.append(0)
    if np.all(np.isinf(a[-1])):
        d.append(-1)

    return np.delete(a, d, axis=0)


def segment_intersection(a: SegmentLike, b: SegmentLike) -> Segment:
    """

    :param a: T-value segments. `Array[float, [A, 2]]`
    :param b: T-value segments. `Array[float, [B, 2]]`
    :return: T-value segments. `Array[float, [C, 2]]`
    """
    a = segment_flatten(a)
    b = segment_flatten(b)
    ret = []

    i = 0
    j = 0
    while i < len(a) and j < len(b):
        p, q = a[i]
        s, t = b[j]
        if p <= s:
            if s < q:  # p <= s < {q,t}
                ret.append((s, min(q, t)))
        else:
            if p < t:  # s < p < {q,t}
                ret.append((p, min(q, t)))

        if q <= t:
            i += 1
        else:
            j += 1

    if len(ret) == 0:
        return np.zeros((0, 2), a.dtype)

    return np.vstack(ret)


def segment_union(a: SegmentLike, b: SegmentLike, gap: float = 0) -> Segment:
    """

    :param a: T-value segments. `Array[float, [A, 2]]`
    :param b: T-value segments. `Array[float, [B, 2]]`
    :param gap: T value
    :return: T-value segments. `Array[float, [C, 2]]`
    """
    a = segment_flatten(a)
    b = segment_flatten(b)

    if len(a) == 0:
        return b
    elif len(b) == 0:
        return a

    ret = []

    i = 0
    j = 0
    while i < len(a) and j < len(b):
        p, q = a[i]
        s, t = b[j]
        if q + gap < s:  # p < q+g < s < t
            ret.append((p, q))
            i += 1
        elif t + gap < p:  # s < t+g < p < q
            ret.append((s, t))
            j += 1
        elif p <= s:  # p < s < {q, t}
            ret.append((p, max(q, t)))
            i += 1
        else:  # s < p < {q, t}
            ret.append((s, max(q, t)))
            j += 1

    while i < len(a):
        ret.append(tuple(a[i]))
        i += 1
    while j < len(b):
        ret.append(tuple(b[j]))
        j += 1
    return segment_flatten(np.vstack(ret))


def segment_diff(a: SegmentLike, b: SegmentLike) -> Segment:
    """

    :param a: T-value segments. `Array[float, [A, 2]]`
    :param b: T-value segments. `Array[float, [B, 2]]`
    :return: T-value segments. `Array[float, [C, 2]]`
    """
    a = segment_flatten(a)
    b = segment_flatten(b)
    return segment_intersection(a, _segment_invert(b))


def segment_contains(segs: SegmentLike, t: np.ndarray) -> np.ndarray:
    """
    whether *t* in the segments.

    :param segs: T-value segments. `Array[float, [N, 2]]`
    :param t: T-value array. `Array[float, R]`
    :return: bool array. `Array[bool, R]`
    """
    segs = segment_flatten(segs)
    x1 = np.less_equal.outer(segs[:, 0], t)  # (N, T), s[0] <= t
    x2 = np.greater_equal.outer(segs[:, 1], t)  # (N, T), s[1] >= t
    xx = np.logical_and(x1, x2)
    return np.logical_or.reduce(xx, axis=0)


def segment_index(segs: SegmentLike, t: np.ndarray) -> np.ndarray:
    """find the index of *segs* where *t* located

    ::

        segs : ---[---]---[---)[---]---
        ret  :  -1  0   -2  1    2   -4

    The following code are true

    .. code-block :: python

        I = segment_index(S, T)
        for i in np.nonzero(I >= 0)[0]:
            assert S[I[i], 0] <= t[i] <= S[I[i], 1]

        for i in np.nonzero(I < 0)[0]:
            if -I[t] - 1 == 0:
                assert t[i] < S[-I[t] - 1, 0]
            elif -I[t] - 1 == len(S):
                assert S[-I[i] - 2, 1] < t[i]
            else:
                assert S[-I[i] - 2, 1] < t[i] < S[-I[t] - 1, 0]

    :param segs: (N, 2) T-value segments
    :param t: (R,) T-value array
    :return: (R,) N-value index array
    :raise ValueError: *segs* has overlapped segments or not sorted
    """
    if len(segment_flatten(segs, closed=False)) != len(segs):
        raise ValueError()

    f = segs.ravel()
    if not is_sorted(f):
        raise ValueError()

    i = np.searchsorted(f, t, 'right')  # Array[int, T]
    pos = i % 2 == 1
    neg = i % 2 == 0
    i[pos] = (i[pos] - 1) // 2
    i[neg] = -i[neg] // 2 - 1
    return i


def segment_overlap(segs: SegmentLike, t: Segment, mode: Literal['in', 'out', 'overlap']) -> np.ndarray:
    """
    * mode == 'in' ::

        returns = [∃ s in S st. t ⊂ s | for t in T]

    * mode == 'out' ::

        returns = [∃ s in S st. s ⊂ t | for t in T]

    * mode == 'overlap' ::

        returns = [∃ s in S st. s ⋂ t ≠ ∅ | for t in T]

    :param segs: (N, 2) T-value segments
    :param t:  (R, 2) T-value segments
    :param mode:
    :return: (R,) bool array
    """
    segs = segment_flatten(segs)
    msk = _segment_overlap(segs, t, mode)
    return np.logical_or.reduce(msk, axis=0)


def segment_overlap_index(segs: SegmentLike, t: Segment, mode: Literal['in', 'out', 'overlap']) -> np.ndarray:
    """
    * mode == 'in' (t is smaller) ::

        return[t] = [∃ i in |S| st. t ⊂ s[i]] for t in T, otherwise [-1]

    * mode == 'out' (t is larger)::

         return[t] = [∃ i in |S| st. s[i] ⊂ t] for t in T, otherwise [-1]

    * mode == 'overlap' ::

       return[t] = [∃ i in |S| st. s[i] ⋂ t ≠ ∅] for t in T, otherwise [-1]

    returns = [min(return[T]), max(return[T])]

    :param segs: (N, 2) T-value segments
    :param t:  (R, 2) T-value segments
    :param mode:
    :return: (2, R) N-value index array
    """
    msk = _segment_overlap(segs, t, mode)
    n, t = msk.shape
    low = np.full_like(msk, n, dtype=int)
    hig = np.full_like(msk, -1, dtype=int)
    grd = np.meshgrid(np.arange(t), np.arange(n))[1]
    low[msk] = grd[msk]
    hig[msk] = grd[msk]
    low = np.min(low, axis=0)
    hig = np.max(hig, axis=0)
    low[low == n] = -1
    return np.vstack([low, hig])


def _segment_overlap(segs: Segment, t: Segment, mode: Literal['in', 'out', 'overlap']) -> np.ndarray:
    if mode == 'in':
        t1 = np.less_equal.outer(segs[:, 0], t[:, 0])  # (N, T), s[0] <= t[0]
        t2 = np.greater_equal.outer(segs[:, 1], t[:, 1])  # (N, T), s[1] >= t[1]
        return np.logical_and(t1, t2)  # (N, T)

    elif mode == 'out':
        t1 = np.greater_equal.outer(segs[:, 0], t[:, 0])  # (N, T), s[0] >= t[0]
        t2 = np.less_equal.outer(segs[:, 1], t[:, 1])  # (N, T), s[1] <= t[1]
        return np.logical_and(t1, t2)

    elif mode == 'overlap':
        t1 = np.greater_equal.outer(segs[:, 0], t[:, 1])  # (N, T), s[0] >= t[1]
        t2 = np.less_equal.outer(segs[:, 1], t[:, 0])  # (N, T), s[1] <= t[0]
        return ~np.logical_or(t1, t2)
    else:
        raise ValueError()


def segment_join(segs: SegmentLike, gap: float = 0) -> Segment:
    """

    :param segs: (N, 2) T-value segments
    :param gap: T
    :return: (N, 2) T-value segments
    """
    segs = segment_flatten(segs)
    if len(segs) <= 1:
        return segs
    ret = segs.copy()
    ret[:, 1] += gap
    ret = segment_flatten(ret)
    ret[:, 1] -= gap
    return ret


def segment_join_index(segs: Segment, gap: float = 0) -> SegmentGroup:
    """

    :param segs: (N, 2) T-value sorted segments
    :param gap: T gap
    :return: (N,) index grouping array
    """
    if segs.ndim != 2 or segs.shape[1] != 2:
        raise ValueError(f'not a (N, 2) segment array : {segs.shape}')
    elif len(segs) == 0:
        return np.array([])
    elif len(segs) == 1:
        return np.array([0])
    elif not is_sorted(segs[:, 0]):
        raise RuntimeError('segments is not sorted')

    t1 = segs[:, 0]
    t2 = segs[:, 1] + gap
    t3 = np.sign(np.diff(t2, prepend=t2[0]))
    t3[t3 < 0] = 0
    t4 = np.cumsum(np.sign(t3).astype(int))
    for t in range(1, np.max(t4) + 1):
        sx = t4 == t - 1
        s1 = np.max(t2[sx])
        s2 = t1[t4 == t][0]
        if s2 <= s1:
            t4[sx] = t
    return t4


def segment_map(f: Callable[[np.ndarray], float],
                segs: SegmentLike,
                t: np.ndarray,
                v: np.ndarray = None) -> np.ndarray:
    """

    :param f: function ((N,) V-value array) -> R-value
    :param segs: (S, 2) T-value segment
    :param t: (T,) T-value array
    :param v: (T,) V-value array. If `None`, use *t*.
    :return: (S,) R-value array
    """
    if v is None:
        v = t
    return segment_group_map(f, segment_index(as_segment(segs), t), v)


def segment_group_map(f: Callable[[np.ndarray], float],
                      group: SegmentGroup,
                      v: np.ndarray) -> np.ndarray:
    """

    :param f: function ((N,) V-value array) -> R-value
    :param group: (T) S-group array. Only non-negative value will be considered.
    :param v: (T,) T-value array
    :return: (S,) R-value array
    """
    ret = []
    for i in range(np.max(group) + 1):
        x = np.nonzero(group == i)[0]
        ret.append(f(v[x]))

    return np.array(ret)


def segment_sample(segs: SegmentLike) -> '_SegmentSampleHelper':
    return _SegmentSampleHelper(segment_flatten(segs))


class _SegmentSampleHelper:
    def __init__(self, segs: Segment):
        self.segs = segs

    def random(self, time_duration: float, sample_times: int) -> Segment:
        if time_duration < 0:
            raise ValueError()
        if sample_times < 0:
            raise ValueError()
        if sample_times == 0:
            return np.zeros((0, 2), dtype=float)

        segs = self.segs[segment_duration(self.segs) >= time_duration]
        cum = np.concatenate([[0], np.cumsum(segment_duration(segs) - time_duration)])
        total = cum[-1]
        a = np.sort(np.random.random(sample_times) * total)
        i = np.searchsorted(cum, a, side='left') - 1

        ret = np.zeros((sample_times, 2), dtype=float)
        ret[:, 0] = a - cum[i] + segs[i, 0]
        ret[:, 1] = ret[:, 0] + time_duration
        return ret

    def uniform(self, time_duration: float, sample_times: int = None) -> Segment:
        if time_duration < 0:
            raise ValueError()

        if sample_times is not None:
            if sample_times < 0:
                raise ValueError()
            if sample_times == 0:
                return np.zeros((0, 2), dtype=float)

        segs = self.segs[segment_duration(self.segs) >= time_duration]
        count = (segment_duration(segs) / time_duration).astype(int)
        total = np.sum(count)
        if sample_times is None:
            sample_times = total
        if total < sample_times:
            raise RuntimeError('d*t larger than dur(segs)')

        cum = np.concatenate([[0], np.cumsum(count)])

        a = np.sort(np.random.choice(np.arange(cum[-1]), sample_times, replace=False))

        i = np.searchsorted(cum, a, side='right') - 1

        ret = np.zeros((sample_times, 2), dtype=float)
        for j in np.unique(i):
            k = np.nonzero(i == j)[0]
            n = len(k)
            assert n <= count[j], f'{n=} > {count[j]}'
            seg = segs[j]
            dur = (seg[1] - seg[0]) - time_duration * n
            assert dur >= 0
            r = np.sort(np.random.random(n)) * dur
            ret[k, 0] = seg[0] + r + time_duration * np.arange(n)

        ret[:, 1] = ret[:, 0] + time_duration

        return ret

    def bins(self, time_duration: float, sample_times: int = None, interval: float = 0) -> Segment:
        return segment_bins(self.segs, time_duration, interval, sample_times)


def segment_bins(segs: SegmentLike,
                 duration: float,
                 interval: float = 0,
                 nbins: int = None) -> Segment:
    """
    Divide *segs* into equal-size sub-segments with equal *duration* and equal *interval*.

    ::

        returns = [(start := R[0] + (i+d)*j, start + d)] ⊆ segs, for j in [0, t)

    :param segs: (N, 2) T-value segment
    :param duration: T value
    :param interval: T value
    :param nbins: number of bins
    :return: (*nbins*, 2) T-value segment
    """
    if duration < 0:
        raise ValueError()

    if interval < 0:
        raise ValueError()

    if nbins is not None:
        if nbins < 0:
            raise ValueError()
        if nbins == 0:
            return np.zeros((0, 2), dtype=float)

    segs = segment_flatten(segs)

    # number of bins per segments
    count = ((interval + segment_duration(segs)) / (duration + interval)).astype(int)

    # maximal total bins
    total = np.sum(count)

    if nbins is None:
        nbins = total
    else:
        nbins = min(total, nbins)

    ret = np.zeros((nbins, 2), dtype=float)
    p = 0
    for j in range(len(count)):
        n = min(int(count[j]), nbins - p)
        k = p + np.arange(n)
        ret[k, 0] = segs[j, 0] + (duration + interval) * np.arange(n)
        p += n
        if p >= nbins:
            break

    ret[:, 1] = ret[:, 0] + duration
    return ret


def shuffle_time(t: np.ndarray,
                 method: Callable[[np.ndarray], np.ndarray],
                 segs: Segment = None,
                 duration: float = np.inf,
                 circular=True):
    """
    Shuffle *t* by remapping function *method* for *t* in *segs*.

    :param t: (N,) T-value array
    :param method: time remapping function with signature ((K,) T-value array) -> (K,) T-value array
    :param segs: (S, 2) segment, only shift time in the segments.
    :param duration: The maximal T value
    :param circular: keep total event number. If epochs is given, keep total event number in epochs.
    :return: (N',) T-value array
    """
    if segs is None:
        ret = method(t)

        if not np.isinf(duration):
            if circular:
                while np.any(x := ret > duration):
                    ret[x] -= duration
            else:
                ret = np.delete(ret, ret > duration, 0)
    else:
        ret = _shuffle_time_in_segment(t, method, segs, (0, duration), circular=circular)

    return np.sort(ret)


def shuffle_time_uniform(t: np.ndarray,
                         segs: Segment = None,
                         *,
                         duration: float = np.inf,
                         circular=True) -> np.ndarray:
    """
    Shuffle *t* for *t* in *segs*.

    :param t: (N,) T-value array
    :param segs: (S, 2) segment, only shift time in the segments.
    :param duration: The maximal T value
    :param circular: keep total event number. If epochs is given, keep total event number in epochs.
    :return: (N',) T-value array
    """
    if np.isinf(duration):
        duration = np.max(t)

    def method(it: np.ndarray) -> np.ndarray:
        return np.random.uniform(0, duration, size=it.shape)

    return shuffle_time(t, method, segs, duration, circular)


def shuffle_time_normal(t: np.ndarray,
                        loc: float = 0,
                        scale: float = 1,
                        segs: Segment = None,
                        *,
                        duration: float = np.inf,
                        circular=True) -> np.ndarray:
    """
    Shuffle *t* with add a value from a normal distribution for *t* in *segs*.

    :param t: (N,) T-value array
    :param loc: mean of the normal distribution
    :param scale: std of the normal distribution
    :param segs: (S, 2) segment, only shift time in the segments.
    :param duration: The maximal T value
    :param circular: keep total event number. If epochs is given, keep total event number in epochs.
    :return: (N',) T-value array
    """

    def method(it: np.ndarray) -> np.ndarray:
        return it + np.random.normal(loc, scale, size=it.shape)

    return shuffle_time(t, method, segs, duration, circular)


def shift_time(t: np.ndarray,
               shift: float,
               segs: Segment = None,
               *,
               duration: float = np.inf,
               circular=True):
    """
    Shift *t* with a *shift* value for *t* in *segs*.

    :param t: (N,) T-value array
    :param shift: shift T value, positive value.
    :param segs:  (S, 2) segment, only shift time in the segments.
    :param duration: The maximal T value for wrapping when *circular*
    :param circular: keep total event number. If epochs is given, keep total event number in epochs.
    :return: (N',) T-value array
    """
    if circular:
        if not (0 <= shift <= duration):
            raise ValueError(f'illegal {shift=}')

    def method(it: np.ndarray) -> np.ndarray:
        return it + shift

    return shuffle_time(t, method, segs, duration, circular)


def _shuffle_time_in_segment(t: np.ndarray,
                             shift: Callable[[np.ndarray], np.ndarray],
                             epochs: Segment,
                             duration: tuple[float, float],
                             circular=True):
    epochs = segment_flatten(epochs)  # Array[float, E', 2]

    ret = t.astype(float, copy=True)  # Array[float, T]
    i = segment_index(epochs, t)  # Array[E', T], index of seg for all t
    x = np.nonzero(i >= 0)[0]  # Array[T, T'], index of t for valid t
    i = i[x]  # Array[E', T'], index of seg for valid t
    sd = segment_duration(epochs)  # Array[float, E']
    sd = np.concatenate([[0], np.cumsum(sd)])  # Array[float, E' + 1]
    ret[x] = shift(ret[x] - epochs[i, 0] + sd[i])

    if circular:
        ret[x] %= sd[-1]
    else:
        o = x[ret[x] > duration[1]]
        ret[o] = np.nan
        o = x[ret[x] < duration[0]]
        ret[o] = np.nan
        x = x[~np.isnan(ret[x])]

    i = np.searchsorted(sd, ret[x], side='right') - 1  # Array[E', T'], new index of seg for shifted t
    ret[x] = ret[x] + epochs[i, 0] - sd[i]

    if not circular:
        ret = ret[~np.isnan(ret)]

    return ret


def foreach_map(v: np.ndarray,
                f: Callable[[np.ndarray], np.ndarray],
                indices_or_sections: np.ndarray) -> np.ndarray:
    """
    Map function ``f`` into vector ``v`` in given sections.

    :param v: Input 1d array
    :param f: Function to apply to each segment
    :param indices_or_sections: If indices_or_sections is an integer, N, the array will be divided into N equal arrays along axis.
        If such a split is not possible, an error is raised
    :return:
    """
    if v.ndim != 1 or indices_or_sections.ndim != 1:
        raise RuntimeError('')

    split_list = np.split(v, indices_or_sections)

    return np.array(list(map(f, split_list[:-1])))  # avoid empty list if divisible


def grouped_iter(v: ArrayLike | Iterable, n: int) -> zip:
    """
    Groups elements from the input iterable ``v`` into tuples of length ``n``

    >>> list(grouped_iter([1, 2, 3, 4, 5, 6], 2))
    [(1, 2), (3, 4), (5, 6)]

    :param v: input iterable to be grouped.
    :param n: number of elements per group
    :return: An iterator over tuples of length n
    """
    return zip(*[iter(v)] * n)



"""
Locomotion
============

This module provide usage for selected locomotion epoch
Including running/stationary, remove jumping frames, interpolation of gaps...


"""

from .epoch import *
from .position import *



from __future__ import annotations

from pathlib import Path
from typing import Literal, overload

import numpy as np
import polars as pl
from brainglobe_atlasapi import BrainGlobeAtlas
from tqdm import tqdm

from neuralib.io import save_json, load_json
from neuralib.io.core import ATLAS_CACHE_DIRECTORY
from neuralib.typing import PathLike
from neuralib.util.tqdm import download_with_tqdm
from neuralib.util.utils import ensure_dir
from neuralib.util.verbose import print_save, print_load

__all__ = [
    'ATLAS_NAME',
    'load_bg_structure_tree',
    'load_bg_volumes',
    'get_children',
    'get_annotation_ids',
    'get_leaf_in_annotation',
    'build_annotation_leaf_map',
    #
    'get_dorsal_cortex'
]

ATLAS_NAME = Literal[
    'allen_mouse_10um',
    'allen_mouse_25um',
    'allen_mouse_50um',
    'allen_mouse_100um',
    'kim_mouse_10um',
    'kim_mouse_25um',
    'kim_mouse_50um',
    'kim_mouse_100um',
    'perens_lsfm_mouse_20um',
    'perens_stereotaxic_mouse_mri_25um',
    'princeton_mouse_20um',
]
"""Atlas Name From BrainGlobeAtlas"""


def load_bg_structure_tree(atlas_name: ATLAS_NAME = 'allen_mouse_10um', *,
                           check_latest: bool = True,
                           paired: bool = False) -> pl.DataFrame:
    """
    Load structure dataframe or dict from `brainglobe_atlasapi`

    :param atlas_name: :attr:`~neuralib.atlas.data.ATLAS_NAME`
    :param check_latest: If check the brainglobe api latest version
    :param paired: To only ``acronym`` & ``parent_acronym`` fields
    :return:
    """
    file = BrainGlobeAtlas(atlas_name, check_latest=check_latest).root_dir / 'structures.csv'
    df = pl.read_csv(file).with_columns(pl.col('parent_structure_id').cast(pl.Int64))
    df = df.join(
        df.select([pl.col("id").alias("parent_structure_id"), pl.col("acronym").alias("parent_acronym")]),
        on="parent_structure_id",
        how="left"
    )

    if paired:
        name = df.select(pl.col('acronym'), pl.col('id'), pl.col('parent_structure_id'))
        join_df = name.join(name, left_on='parent_structure_id', right_on='id')
        parent_child = join_df.select(pl.col('acronym'), pl.col('acronym_right').alias('parent_acronym'))
        return parent_child
    else:
        return df


def load_bg_volumes(atlas_name: ATLAS_NAME = 'allen_mouse_10um',
                    cached_file: PathLike | None = None,
                    force: bool = False) -> pl.DataFrame:
    """
    Load structure tree dataframe with volume for each region ``volume_mm3`` ::

        ┌─────────┬─────┬─────────────────┬─────────────────┬────────────────┬────────────────┬────────────┐
        │ acronym ┆ id  ┆ name            ┆ structure_id_pa ┆ parent_structu ┆ parent_acronym ┆ volume_mm3 │
        │ ---     ┆ --- ┆ ---             ┆ th              ┆ re_id          ┆ ---            ┆ ---        │
        │ str     ┆ i64 ┆ str             ┆ ---             ┆ ---            ┆ str            ┆ f64        │
        │         ┆     ┆                 ┆ str             ┆ i64            ┆                ┆            │
        ╞═════════╪═════╪═════════════════╪═════════════════╪════════════════╪════════════════╪════════════╡
        │ VI      ┆ 653 ┆ Abducens        ┆ /997/8/343/1065 ┆ 370            ┆ MY-mot         ┆ 0.030332   │
        │         ┆     ┆ nucleus         ┆ /354/370/653/   ┆                ┆                ┆            │
        │ AOB     ┆ 151 ┆ Accessory       ┆ /997/8/567/688/ ┆ 698            ┆ OLF            ┆ 0.652032   │
        │         ┆     ┆ olfactory bulb  ┆ 695/698/151/    ┆                ┆                ┆            │
        │ …       ┆ …   ┆ …               ┆ …               ┆ …              ┆ …              ┆ …          │
        │ von     ┆ 949 ┆ vomeronasal     ┆ /997/1009/967/9 ┆ 967            ┆ cm             ┆ 0.013428   │
        │         ┆     ┆ nerve           ┆ 49/             ┆                ┆                ┆            │
        └─────────┴─────┴─────────────────┴─────────────────┴────────────────┴────────────────┴────────────┘

    :param atlas_name: :attr:`~neuralib.atlas.data.ATLAS_NAME`
    :param cached_file: Cached file path.
    :param force: Force overwrite the cached file
    :return:
    """
    if cached_file is None:
        cached_file = ensure_dir(ATLAS_CACHE_DIRECTORY) / f'{atlas_name}_bg_volumes.csv'

    if cached_file.exists() and not force:
        print_load(cached_file)
        return pl.read_csv(cached_file)
    else:
        df = load_bg_structure_tree(atlas_name)
        lut = build_annotation_leaf_map(atlas_name)
        bg = BrainGlobeAtlas(atlas_name)
        flat_annotation = bg.annotation.ravel()
        voxel_volume_mm3 = (bg.resolution[0] / 1000) ** 3

        volumes = []
        ids = df['id'].to_list()

        for region_id in tqdm(ids, desc=f"calculating volumes and save cache ({atlas_name})"):
            leaf_ids = lut.get(region_id, [])

            if not leaf_ids:
                volumes.append(-1)  # if not found
                continue

            mask = np.isin(flat_annotation, leaf_ids)
            count = np.count_nonzero(mask)
            vol_mm3 = count * voxel_volume_mm3
            volumes.append(vol_mm3)

        ret = df.with_columns(pl.Series(name='volume_mm3', values=volumes))
        ret.write_csv(cached_file)
        print_save(cached_file)

        return ret


@overload
def get_children(parent: int, *,
                 dataframe: bool = False,
                 atlas_name: ATLAS_NAME = 'allen_mouse_10um') -> list[int] | pl.DataFrame:
    pass


@overload
def get_children(parent: str, *,
                 dataframe: bool = False,
                 atlas_name: ATLAS_NAME = 'allen_mouse_10um') -> list[str] | pl.DataFrame:
    pass


def get_children(parent: int | str, *,
                 dataframe: bool = False,
                 atlas_name: ATLAS_NAME = 'allen_mouse_10um') -> list[str] | pl.DataFrame:
    """
    Get children brain region id or acronym from its parent

    :param parent: id or acronym
    :param dataframe: return as dataframe, otherwise return as list
    :param atlas_name: :attr:`~neuralib.atlas.data.ATLAS_NAME`
    :return:
    """
    df = load_bg_structure_tree(atlas_name=atlas_name)
    return _get_children(df, parent, dataframe)


def _get_children(df, parent, dataframe):
    if isinstance(parent, int):
        ret = df.filter(pl.col('parent_structure_id') == parent)
        field = 'id'
    elif isinstance(parent, str):
        ret = df.filter(pl.col('parent_acronym') == parent)
        field = 'acronym'
    else:
        raise TypeError('')

    if not dataframe:
        ret = ret[field].to_list()

    return ret


# ============= #
# BG Annotation #
# ============= #

def get_annotation_ids(atlas_name: ATLAS_NAME = 'allen_mouse_10um', check_latest: bool = True) -> np.ndarray:
    """
    Get unique annotation id

    :param atlas_name: :attr:`~neuralib.atlas.data.ATLAS_NAME`
    :param check_latest:
    :return:
    """
    annotation = BrainGlobeAtlas(atlas_name, check_latest=check_latest).annotation
    return np.unique(annotation)


def get_leaf_in_annotation(region: int | str, *,
                           name: bool = False,
                           cached_file: PathLike | None = None,
                           atlas_name: ATLAS_NAME = 'allen_mouse_10um') -> list[int] | list[str]:
    """
    Get a list of annotation {id, acronym} with given region {id, acronym}

    :param region: Region id or region acronym
    :param name: If True, return acronym, otherwise return id
    :param cached_file: Cached json for the annotation_leaf_map
    :param atlas_name: :attr:`~neuralib.atlas.data.ATLAS_NAME`
    :return: List of annotation {id, acronym}
    """
    tree = load_bg_structure_tree(atlas_name=atlas_name)

    # to id
    if isinstance(region, str):
        region_ids = tree.filter(pl.col('acronym') == region)['id'].to_list()
        if len(region_ids) != 1:
            raise RuntimeError(f"The region {region} is not a valid acronym")
        region = region_ids[0]

    dy = build_annotation_leaf_map(cached_file=cached_file)

    try:
        result = dy[region]
    except KeyError:
        raise ValueError(f'Invalid region: {region}')

    if name:
        result = tree.filter(pl.col('id').is_in(result))['acronym'].to_list()

    return result


def build_annotation_leaf_map(atlas_name: ATLAS_NAME = 'allen_mouse_10um', *,
                              cached_file: PathLike | None = None,
                              force: bool = False) -> dict[int, list[int]]:
    """
    Get all region id (key) and list of annotation id (values)

    :param atlas_name: :attr:`~neuralib.atlas.data.ATLAS_NAME`
    :param cached_file: Cached json file path
    :param force: Force re-compute the cached file
    :return:
    """
    if cached_file is None:
        cached_file = ensure_dir(ATLAS_CACHE_DIRECTORY) / f'{atlas_name}_annotation_leaf.json'

    if Path(cached_file).suffix != '.json':
        raise ValueError('not a json file')

    #
    if cached_file.exists() and not force:
        data = load_json(cached_file, verbose=False)
        leaf_map = {int(k): v for k, v in data.items()}
    else:
        tree = load_bg_structure_tree(atlas_name)
        id_to_children = _build_id_to_children_map(tree)
        annotation_ids = set(get_annotation_ids())

        leaf_map = {}

        def collect(rid):
            if rid in leaf_map:
                return leaf_map[rid]
            if rid in annotation_ids:
                leaf_map[rid] = [rid]
            else:
                result = []
                for child in id_to_children.get(rid, []):
                    result.extend(collect(child))
                leaf_map[rid] = result
            return leaf_map[rid]

        all_ids = tree['id'].to_list()
        for rid in all_ids:
            collect(rid)

        save_json(cached_file, leaf_map)

    return leaf_map


def _build_id_to_children_map(tree: pl.DataFrame) -> dict[int, list[int]]:
    df = tree.select(['id', 'parent_structure_id'])
    grouped = df.group_by('parent_structure_id', maintain_order=False).agg(pl.col('id'))
    return {row['parent_structure_id']: row['id'] for row in grouped.iter_rows(named=True)}


# =============== #
# Allen Resources #
# =============== #

def get_dorsal_cortex(output_dir: Path | None = None) -> Path:
    """
    Get example dorsal projection annotation svg file

    .. seealso::

        https://community.brain-map.org/t/aligning-dorsal-projection-of-mouse-common-coordinate-framework-with-wide-field-images-of-mouse-brain/140/2

    :param output_dir: Output directory for caching
    :return: Output file path
    """

    if output_dir is None:
        output_dir = ensure_dir(ATLAS_CACHE_DIRECTORY)

    filename = 'cortical_map_top_down.svg'
    output = output_dir / filename

    if not output.exists():
        url = 'http://connectivity.brain-map.org/assets/cortical_map_top_down.svg'
        content = download_with_tqdm(url)

        with open(output, 'wb') as f:
            f.write(content.getvalue())
            print_save(output, verb='DOWNLOAD')

    return output



from typing import Optional, NamedTuple, ClassVar

from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from neuralib.typing import PathLike

__all__ = ['VennHandler',
           'VennDiagram']


class VennHandler(NamedTuple):
    subset_a: int
    """whole number with condition a"""
    subset_b: int
    """whole number with condition b"""
    subset_overlap: int
    """whole number with condition a & b"""

    total_set: Optional[int] = None

    @property
    def chance_level(self) -> float:
        if self.total_set is None:
            raise ValueError('call with_total first')

        fa = self.subset_a / self.total_set
        fb = self.subset_b / self.total_set

        return fa * fb * 100

    def with_total(self, total: int) -> 'VennHandler':
        """total set number. should include the non-classified population"""
        return self._replace(total_set=total)

    def get_pure_number(self) -> tuple[int, ...]:
        a = self.subset_a - self.subset_overlap
        b = self.subset_b - self.subset_overlap
        return tuple([a, b, self.subset_overlap])

    def get_pure_fraction(self) -> tuple[float, ...]:
        a, b, o = self.get_pure_number()
        return tuple([a / self.total_set * 100,
                      b / self.total_set * 100,
                      o / self.total_set * 100])


class VennDiagram:
    DEFAULT_COLORS: ClassVar[tuple[str, ...]] = ('r', 'g', 'b')

    def __init__(self,
                 subsets: dict[str, int],
                 *,
                 colors: tuple[str, ...] | None = None,
                 ax: Axes | None = None,
                 **kwargs):
        """

        :param subsets: Dictionary of set label and its value
        :param colors: colors of each venn
        :param ax: ``Axes``
        :param kwargs: additional args passed to ``matplotlib_venn.venn2()`` or ``matplotlib_venn.venn3()``
        """
        self.subsets = subsets

        self.total: int | None = None
        self.contain_intersection: bool = False
        self._intersections: dict[str, int] = {}

        # fig
        if colors is not None and len(colors) != len(self):
            raise ValueError('length of colors need to be the same as length of subset label')
        self.colors = colors or VennDiagram.DEFAULT_COLORS
        self.ax = ax

        self.kwargs = kwargs

    def __len__(self):
        return len(self.subsets)

    @property
    def intersections(self) -> dict[str, int]:
        """intersection for sets"""
        return self._intersections

    @property
    def max_intersection_areas(self):
        """maximal number of intersection areas"""
        if len(self) == 2:
            return 1
        elif len(self) == 3:
            return 4
        else:
            raise NotImplementedError('')

    @property
    def labels(self) -> tuple[str, ...]:
        """set names"""
        return tuple(self.subsets.keys())

    @property
    def subsets_percentage(self) -> dict[str, float]:
        """percentage of each subset"""
        if self.total is None:
            raise RuntimeError('add total first')

        source = {
            k: round(v / self.total * 100, 2)
            for k, v in self.subsets.items()
        }

        inter = {
            k: round(v / self.total * 100, 2)
            for k, v in self.intersections.items()
        }

        return {**source, **inter}

    def add_total(self, value: int):
        """
        Add total value to the venn diagram
        :param value: value to be added
        """
        self.total = value

    def add_intersection(self, group: str, value: int):
        """
        Add intersection values using "&"

        :param group: i.e., `a & b`
        :param value: value of the intersection
        """

        src = [g.strip() for g in group.split('&')]

        for inter in list(self.intersections.keys()):
            if all(g in inter for g in src):
                raise ValueError('intersection value already existed')

        for g in src:
            if g not in self.subsets:
                raise ValueError(f"Set '{g}' does not exist in the subsets.")

        self._intersections['&'.join(src)] = value

    def get_chance_level(self, *label) -> float:
        if self.contain_intersection:
            raise RuntimeError('chance level should not contain intersection')

        if self.total is None:
            raise RuntimeError('add total first')

        x = 100
        for it in label:
            x *= self.subsets[it] / self.total

        return x

    def get_intersection(self, *label: str) -> int:
        """
        Get intersection value from labels

        :param label:i.e., `a & b`
        :return: intersection value
        """
        k = '&'.join([*label])
        return self.intersections.get(k, 0)

    def with_intersection(self):
        """Add intersection value into subsets"""
        if self.contain_intersection:
            raise RuntimeError('already contain intersection')

        ret = {}
        if len(self) == 2:
            inter = self.get_intersection(*self.labels)
            for k, v in self.subsets.items():
                ret[k] = v + inter
        elif len(self) == 3:
            ab = self.get_intersection(*self.labels[:2])
            ac = self.get_intersection(self.labels[0], self.labels[2])
            bc = self.get_intersection(*self.labels[1:])
            abc = self.get_intersection(*self.labels)
            for i, (k, v) in enumerate(self.subsets.items()):
                if i == 0:
                    ret[k] = v + ab + ac + abc
                elif i == 1:
                    ret[k] = v + ab + bc + abc
                elif i == 2:
                    ret[k] = v + ac + bc + abc
        else:
            raise NotImplementedError('')

        self.subsets = ret
        self.contain_intersection = True

    # ================ #
    # Plotting Methods #
    # ================ #

    def plot(self, add_title: bool = True):
        """
        Plot the venn diagram

        :param add_title: Add percentage information and total as title
        """
        if self.ax is None:
            self.ax = plt.gca()

        n_subsets = len(self)
        if n_subsets == 2:
            self._venn2()
        elif n_subsets == 3:
            self._venn3()

        if add_title and self.total is not None:
            self.ax.set_title(self.title)

        self.ax.set_axis_off()
        self.ax.set_xticks([])
        self.ax.set_yticks([])

    @staticmethod
    def show():
        """Show figure"""
        plt.show()

    @staticmethod
    def savefig(output: PathLike):
        """
        Save figure

        :param output: fig output
        """
        plt.savefig(output)

    @property
    def title(self) -> str:
        """title of the plot"""
        ret = [
            f'percentage: {self.subsets_percentage}%',
            f'total: {self.total}'
        ]

        return '\n'.join(ret)

    # noinspection PyTypeChecker
    def _venn2(self):
        """subsets = (a, b, a&b)"""
        from matplotlib_venn import venn2
        subsets = list(self.subsets.values()) + [self.get_intersection(*self.labels)]

        venn2(subsets=tuple(subsets),
              set_labels=self.labels,
              set_colors=self.colors[:2],
              ax=self.ax,
              **self.kwargs)

    # noinspection PyTypeChecker
    def _venn3(self):
        """subsets = (a, b, a&b, c, a&c, b&c, a&b&c)"""
        from matplotlib_venn import venn3
        v = list(self.subsets.values())
        a = v[0]
        b = v[1]
        c = v[2]
        ab = self.get_intersection(*self.labels[:2])
        ac = self.get_intersection(self.labels[0], self.labels[2])
        bc = self.get_intersection(*self.labels[1:])
        abc = self.get_intersection(*self.labels)

        venn3(subsets=(a, b, ab, c, ac, bc, abc),
              set_labels=self.labels,
              set_colors=self.colors,
              ax=self.ax,
              **self.kwargs)



import itertools

import numpy as np
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.cm import ScalarMappable
from matplotlib.colorbar import ColorbarBase
from mpl_toolkits.axes_grid1 import make_axes_locatable

__all__ = [
    'DiscreteColorMapper',
    'get_customized_cmap',
    'insert_colorbar',
    'insert_cyclic_colorbar'
]


class DiscreteColorMapper:
    """map color to iterable object

    **Example of a ``dict`` palette**

        >>> palette = {3: ('#3182bd', '#6baed6', '#9ecae1'),
        ...            5: ('#a1d99b', '#c7e9c0', '#756bb1', '#9e9ac8', '#bcbddc')}
        >>> cmapper = DiscreteColorMapper(palette, 5)
        >>> x = ['1', '2', '3']
        >>> color_list = [cmapper[i] for i in x]

    **Example of a mpl ``str`` palette**

        >>> cmapper = DiscreteColorMapper('viridis', 20)
        >>> regions = ['rsc', 'vis', 'hpc']
        >>> color_list = [cmapper[r] for r in regions]
    """

    def __init__(self,
                 palette: dict[int, tuple[str, ...]] | str,
                 number: int):
        """

        :param palette: A custom dictionary color palette or a Matplotlib colormap name
        :param number: Number of colors to cycle through in the palette or colormap
        """
        if isinstance(palette, dict):
            if number not in palette:
                raise ValueError(f"Palette dictionary does not contain key '{number}'.")
            self._color_pool = iter(itertools.cycle(palette[number]))
        elif isinstance(palette, str):
            import matplotlib as mpl
            mapper = mpl.colormaps[palette].resampled(number)
            self._color_pool = iter(itertools.cycle(mapper(range(number))))
        else:
            raise TypeError('palette must be a dict or str')

        self._key_pool = {}

    def __getitem__(self, item: str) -> str | np.ndarray:
        """get single color"""
        if item in self._key_pool:
            return self._key_pool[item]
        else:
            color = next(self._color_pool)
            self._key_pool[item] = color
            return color


def get_customized_cmap(name: str,
                        value: tuple[float, float],
                        numbers: int,
                        endpoint: bool = True) -> np.ndarray:
    """
    Generate gradient color map array.
    `N` = number of color

    :param name: name of cmap
    :param value: value range, could be 0-1
    :param numbers: `N`
    :param endpoint: If cyclic colormap, then used `False`
    :return: RGBA. `Array[float, [N, 4]]`
    """
    cmap = plt.get_cmap(name)
    return cmap(np.linspace(*value, numbers, endpoint=endpoint))


def insert_colorbar(ax: Axes, im: ScalarMappable, inset: bool = False, **kwargs) -> ColorbarBase:
    """
    Insert colormap

    :param ax: ``Axes``
    :param im: ``ScalarMappable``
    :param inset: use ``inset axes``, otherwise append a new axis
    :param kwargs: Additional args pass to ``ax.figure.colorbar``
    :return:
    """
    if inset:
        from mpl_toolkits.axes_grid1.inset_locator import inset_axes
        cax = inset_axes(
            ax,
            width="5%",
            height="25%",
            loc='upper left',
            bbox_to_anchor=(1.01, 0., 1, 1),
            bbox_transform=ax.transAxes,
            borderpad=0,
        )
    else:
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="7%", pad=0.1)

    return ax.figure.colorbar(im, cax=cax, **kwargs)


def insert_cyclic_colorbar(ax: Axes,
                           im: ScalarMappable,
                           *,
                           num_colors: int = 12,
                           num_labels: int = 4,
                           width: float = 0.6,
                           inner_diameter: float = 0.6,
                           vmin: float | None = None,
                           vmax: float | None = None) -> None:
    """
    Insert cyclic colormap in ``inset_axes``

    :param ax: ``Axes``
    :param im: ``ScalarMappable``
    :param num_colors: Number of color in the cyclic colorbar
    :param num_labels: Number of labels in the cyclic colorbar
    :param width: Width of the each color
    :param inner_diameter: The size of the inner circle
    :param vmin: Min value of the colormap, equal to ``vmax`` in cyclic data
    :param vmax: Max value of the colormap, equal to ``vmin`` in cyclic data
    """
    polar_ax = ax.inset_axes((1.1, 0.65, 0.4, 0.4), polar=True)
    theta = np.linspace(0, 2 * np.pi, num_colors, endpoint=False)
    r1 = np.ones_like(theta)
    r2 = np.ones_like(theta) * inner_diameter

    cmap = im.cmap.name
    cyclic_cmap = ['twilight', 'twilight_shifted', 'hsv']
    if cmap not in cyclic_cmap:
        raise ValueError(f'cmap should be one of the {cyclic_cmap}, not "{cmap}"')
    colors = get_customized_cmap(cmap, (0, 1), num_colors, endpoint=False)
    polar_ax.bar(theta, r1, color=colors, width=width, bottom=r2)

    # Add labels corresponding to the data values
    angles = np.linspace(0, 2 * np.pi, num_labels, endpoint=False)

    if vmin is None:
        vmin = im.norm.vmin
    if vmax is None:
        vmax = im.norm.vmax
    values = np.linspace(vmin, vmax, num_labels, endpoint=False)

    for angle, value in zip(angles, values):
        polar_ax.text(angle, 3, f'{value:.1f}', horizontalalignment='center', verticalalignment='center')

    polar_ax.set_yticklabels([])
    polar_ax.set_xticks([])  # remove angle labels
    polar_ax.grid(False)  # remove the grid
    polar_ax.set_frame_on(False)



from pathlib import Path

import cv2
import numpy as np
from bokeh.layouts import column
from bokeh.model import Model
from bokeh.models import ColumnDataSource, GlyphRenderer, Slider
from bokeh.plotting import figure
from tifffile import tifffile

from argclz import AbstractParser, argument
from neuralib.dashboard import ViewComponent, Figure, View, BokehServer
from neuralib.imaging.widefield import SequenceFFT
from neuralib.typing import PathLike

__all__ = ['WideFieldFFTViewOption']


class WideFieldFFTViewComponent(ViewComponent):
    source: ColumnDataSource
    render: GlyphRenderer

    def __init__(self, dat: np.ndarray | None):
        self.dat = dat
        self.source = ColumnDataSource(data=dict(x=[], y=[], map=[]))

    def plot(self, fig: Figure, **kwargs):
        self.render = fig.image_rgba(
            image='map',
            x=0,
            y=0,
            dw='x',
            dh='y',
            source=self.source,
            level='image'
        )

        fig.xgrid.grid_line_color = None
        fig.ygrid.grid_line_color = None
        fig.axis.visible = False

    def update(self, **kwargs):
        if (image := self.dat) is not None:
            self.update_image(image)

    def update_image(self, dat: np.ndarray):
        self.dat = dat
        if dat is not None:
            y, x = dat.shape
            self.source.data = dict(x=[x], y=[y], map=[dat])


class WideFieldFFTView(View):
    slider_saturation_factor: Slider
    slider_value_perc: Slider
    slider_saturation_perc: Slider

    fig_map: Figure
    view_map = WideFieldFFTViewComponent

    fft: SequenceFFT

    def __init__(self, seq_path: PathLike):
        self.seq_path = Path(seq_path)
        self.fft = SequenceFFT(tifffile.imread(seq_path))

    def setup(self) -> Model:
        w, h = self.fft.width, self.fft.height

        # slider
        self.slider_saturation_factor = Slider(
            start=0,
            end=1,
            value=0.3,
            step=0.1,
            width=600,
            title='saturation_factor'
        )
        self.slider_saturation_factor.on_change('value', self.on_select_saturation_factor)

        self.slider_value_perc = Slider(
            start=0,
            end=100,
            value=98,
            width=600,
            title='value_perc'
        )

        self.slider_value_perc.on_change('value', self.on_select_value_perc)

        self.slider_saturation_perc = Slider(
            start=0,
            end=100,
            value=90,
            width=600,
            title='saturation_perc'
        )

        self.slider_saturation_perc.on_change('value', self.on_select_saturation_perc)

        # fig
        self.fig_map = figure(title='image_intensity', width=w, height=h)
        self.fig_map = figure(
            title='image_map',
            width=w,
            height=h,
        )

        # view
        self.view_map = WideFieldFFTViewComponent(self.fft.seq)
        self.view_map.plot(self.fig_map)

        return column(
            self.slider_saturation_factor,
            self.slider_value_perc,
            self.slider_saturation_perc,
            self.fig_map
        )

    def on_select_saturation_factor(self, attr: str, old: str, value: str):
        self._on_select_slider_value(attr, old, value)

    def on_select_value_perc(self, attr: str, old: str, value: str):
        self._on_select_slider_value(attr, old, value)

    def on_select_saturation_perc(self, attr: str, old: str, value: str):
        self._on_select_slider_value(attr, old, value)

    def _on_select_slider_value(self, attr: str, old: str, value: str):
        cmap = self.as_colormap(self.slider_saturation_factor.value,
                                self.slider_value_perc.value,
                                self.slider_saturation_perc.value)
        self.run_later(self.view_map.update_image, cmap)

    def as_colormap(self, saturation_factor, value_perc, saturation_perc) -> np.ndarray:
        ret = self.fft.as_colormap(
            saturation_factor=saturation_factor,
            value_perc=value_perc,
            saturation_perc=saturation_perc
        )
        ret = cv2.cvtColor(ret, cv2.COLOR_RGB2RGBA)
        ret = ret.view(dtype=np.uint32).reshape((self.fft.height, self.fft.width))

        return np.flipud(ret)


class WideFieldFFTViewOption(AbstractParser):
    DESCRIPTION = 'View the HSV colormap representation of the Fourier transform results'

    file: str = argument(metavar='FILE', help='file path for the video sequence')

    def run(self):
        server = BokehServer(theme='caliber')
        server.start(WideFieldFFTView(self.file))


if __name__ == '__main__':
    WideFieldFFTViewOption().main()



import napari

from argclz import AbstractParser, argument
from neuralib.io.dataset import load_example_dorsal_cortex

__all__ = ['NapariAlignmentOptions']


class NapariAlignmentOptions(AbstractParser):
    DESCRIPTION = 'View the video sequences and top view of dorsal cortex using napari'

    sequence_path: str = argument(
        metavar="FILE",
        help='file path for the video sequence'
    )

    reference_path: str | None = argument(
        '-R', '--reference',
        default=None,
        help='reference image (e.g., bright-field)'
    )

    dorsal_map_path: str | None = argument(
        '-M', '--map',
        default=None,
        help='dorsal map file, If None then use default'
    )

    color: bool = argument(
        '--color',
        help='alignment map with color',
    )

    def run(self):
        viewer = napari.Viewer()
        viewer.open(self.sequence_path)

        #
        if self.dorsal_map_path is not None:
            viewer.open(self.dorsal_map_path)
        else:
            cortical_map = load_example_dorsal_cortex(color=self.color, cached=True)
            viewer.add_image(cortical_map)

        #
        if self.reference_path is not None:
            viewer.open(self.reference_path)

        napari.run()


if __name__ == '__main__':
    NapariAlignmentOptions().main()




import typing
from typing import TYPE_CHECKING

from .literal import FOREIGN_POLICY, CONFLICT_POLICY

if TYPE_CHECKING:
    pass

__all__ = [
    'CURRENT_DATE', 'CURRENT_TIME', 'CURRENT_TIMESTAMP',
    'PRIMARY', 'UNIQUE', 'unique', 'foreign', 'check',
]

T = typing.TypeVar('T')
missing = object()


@typing.final
class PRIMARY(object):
    """
    annotate a field as a primary key.

    >>> class Example:
    ...     a: Annotated[str, PRIMARY]  # style 1
    """

    def __init__(self, order: typing.Literal['ASC', 'DESC'] = None,
                 conflict: CONFLICT_POLICY = None,
                 auto_increment=False):
        self.order = order
        self.conflict = conflict
        self.auto_increment = auto_increment


@typing.final
class UNIQUE(object):
    """
    annotated a field as a unique key.

    >>> class Example:
    ...     a: Annotated[str, UNIQUE]  # style 1
    """

    def __init__(self, conflict: CONFLICT_POLICY = None):
        self.conflict = conflict


@typing.final
class CURRENT_DATE(object):
    """
    annotated a date field which use current date as its default.

    >>> class Example:
    ...     a: Annotated[datetime.date, CURRENT_DATE]

    """

    def __init__(self):
        raise RuntimeError()


@typing.final
class CURRENT_TIME(object):
    """
    annotated a time field which use current time as its default.

    >>> class Example:
    ...     a: Annotated[datetime.time, CURRENT_TIME]

    """

    def __init__(self):
        raise RuntimeError()


@typing.final
class CURRENT_TIMESTAMP(object):
    """
    annotated a datetime field which use current datetime as its default.

    >>> class Example:
    ...     a: Annotated[datetime.datetime, CURRENT_TIMESTAMP]

    """

    def __init__(self):
        raise RuntimeError()


def unique(conflict: CONFLICT_POLICY = None):
    """
    A decorator to create an unique constraint.
    """

    def _decorator(f):
        setattr(f, '_sql_unique', (conflict,))
        return f

    return _decorator


def foreign(*field,
            update: FOREIGN_POLICY = 'NO ACTION',
            delete: FOREIGN_POLICY = 'NO ACTION'):
    """
    A decorator to create a foreign constraint.

    Common use:

    With a foreign table

    >>> class ForeignTable:
    ...     a: Annotated[str, str]
    ...     b: Annotated[str, str]

    1. mapping one-by-one

    >>> class Example:
    ...     a: Annotated[str, str]
    ...     b: Annotated[str, str]
    ...     @foreign(ForeignTable.a, ForeignTable.b)
    ...     def _foreign(self):
    ...         return self.a, self.b

    2. by default, with the primary keys for the referred foreign table.

    >>> class Example:
    ...     a: Annotated[str, str]
    ...     b: Annotated[str, str]
    ...     @foreign(ForeignTable)
    ...     def _foreign(self):
    ...         return self.a, self.b

    3. Self refered.

    >>> class Example:
    ...     a: Annotated[str, str]
    ...     b: Annotated[str, str]
    ...     @foreign('a')
    ...     def _foreign(self):
    ...         return self.b

    :param field: a foreign table or foreign fields.
    :param update: update policy
    :param delete: delete policy
    """
    if len(field) == 0:
        raise RuntimeError('empty fields')

    def _decorator(f):
        setattr(f, '_sql_foreign', (field, update, delete))
        return f

    return _decorator


def check(field: str = None):
    """
    A decorator to make a check constraint.

    1. by a field

    >>> class Example:
    ...     a: str
    ...     @check('a')
    ...     def check_a(self):
    ...         return self.a != ''

    2. by over all

    >>> class Example:
    ...     a: str
    ...     @check()
    ...     def check_a(self):
    ...         return self.a != ''

    :param field: field name.
    """

    def _decorator(f):
        setattr(f, '_sql_check', (field, ))
        return f

    return _decorator



from argclz.commands import parse_command_args

from .core import BrainRenderCLI
from .probe import ProbeRenderCLI
from .roi import RoiRenderCLI


def main():
    parsers = dict(
        area=BrainRenderCLI,
        roi=RoiRenderCLI,
        probe=ProbeRenderCLI
    )

    parse_command_args(
        parsers=parsers,
        description='BrainRender CLI Options',
        usage="""
        Usage Examples:
    
        Render a brain region:
        $ neuralib_brainrender area -R <REGION, ...>
        $ neuralib_brainrender area -R VISal,VISam,VISl,VISli,VISp,VISpl,VISpm,VISpor
    
        Render a region of interest (ROI) from a file:
        $ neuralib_brainrender roi -F <FILE>
    
        Render a probe placement from a file with depth specification:
        $ neuralib_brainrender probe -F <FILE> --depth <DEPTH>
        """
    )


if __name__ == '__main__':
    main()
